from typing import List

import pytest
import torch
from lightbench.utils import get_optim
from torch import nn

import heavyball
from heavyball.utils import clean, set_torch


class Param(nn.Module):
    def __init__(self, size):
        super().__init__()
        self.weight = nn.Parameter(torch.randn(size))

    def forward(self, inp):
        return self.weight.mean() * inp


@pytest.mark.parametrize("opt", heavyball.__all__)
@pytest.mark.parametrize(
    "size",
    [
        (4, 4, 4, 4),
    ],
)
def test_closure(opt, size: List[int], depth: int = 2, iterations: int = 5, outer_iterations: int = 3):
    clean()
    set_torch()

    opt = getattr(heavyball, opt)

    for _ in range(outer_iterations):
        clean()
        model = nn.Sequential(*[Param(size) for _ in range(depth)]).cuda()
        o = get_optim(opt, model.parameters(), lr=1e-3)

        def _closure():
            loss = model(torch.randn((1, size[0]), device="cuda")).sum()
            loss.backward()
            return loss

        for i in range(iterations):
            o.step(_closure)
            o.zero_grad()
            print(o.state_size())



import functools
import math
from typing import Optional, Type, Union

import torch.optim

from . import chainable as C
from . import utils


class SGD(C.BaseOpt):
    def __init__(
        self,
        params,
        lr=0.0025,
        beta=0.9,
        weight_decay=0,
        warmup_steps=0,
        foreach: bool = True,
        storage_dtype: str = "float32",
        mars: bool = False,
        caution: bool = False,
        mars_gamma: float = 0.0025,
        gradient_clipping: C.str_or_fn = C.use_default,
        update_clipping: C.str_or_fn = C.use_default,
        **kwargs,
    ):
        defaults = locals()
        defaults.pop("self")
        params = defaults.pop("params")
        defaults.update(defaults.pop("kwargs"))

        if kwargs:
            utils.warn_once(f"Working with uncaptured keyword arguments: {kwargs}")

        super().__init__(params, defaults, foreach, gradient_clipping, update_clipping, fns=(C.heavyball_momentum,))


class ForeachAdamW(C.BaseOpt):
    def __init__(
        self,
        params,
        lr=0.0025,
        betas=(0.9, 0.99),
        eps=1e-8,
        weight_decay=0,
        warmup_steps=0,
        foreach: bool = True,
        storage_dtype: str = "float32",
        mars: bool = False,
        caution: bool = False,
        mars_gamma: float = 0.0025,
        gradient_clipping: C.str_or_fn = C.use_default,
        update_clipping: C.str_or_fn = C.use_default,
        palm: bool = C.use_default,
        beta2_scale: float = 0.8,
        **kwargs,
    ):
        defaults = locals()
        defaults.pop("self")
        params = defaults.pop("params")
        defaults.update(defaults.pop("kwargs"))

        if kwargs:
            utils.warn_once(f"Working with uncaptured keyword arguments: {kwargs}")

        super().__init__(params, defaults, foreach, gradient_clipping, update_clipping, palm, fns=(C.update_by_adam,))


class ForeachNAdam(C.BaseOpt):
    def __init__(
        self,
        params,
        lr=0.002,
        betas=(0.9, 0.999),
        eps=1e-8,
        weight_decay=0,
        momentum_decay: float = 4e-3,
        decoupled_weight_decay: bool = False,
        warmup_steps=0,
        foreach: bool = True,
        storage_dtype: str = "float32",
        mars: bool = False,
        caution: bool = False,
        mars_gamma: float = 0.0025,
        gradient_clipping: C.str_or_fn = C.use_default,
        update_clipping: C.str_or_fn = C.use_default,
        palm: bool = C.use_default,
        beta2_scale: float = 0.8,
        **kwargs,
    ):
        defaults = locals()
        defaults.pop("self")
        params = defaults.pop("params")
        defaults.update(defaults.pop("kwargs"))

        if kwargs:
            utils.warn_once(f"Working with uncaptured keyword arguments: {kwargs}")

        super().__init__(params, defaults, foreach, gradient_clipping, update_clipping, palm, fns=(C.update_by_nadam,))


class ForeachAdEMAMix(C.BaseOpt):
    def __init__(
        self,
        params,
        lr=0.001,
        betas=(0.9, 0.999, 0.9999),
        eps=1e-8,
        weight_decay=0,
        alpha: float = 2.0,
        beta3_warmup: Optional[int] = None,
        alpha_warmup: Optional[int] = None,
        warmup_steps=0,
        foreach: bool = True,
        storage_dtype: str = "float32",
        mars: bool = False,
        caution: bool = False,
        mars_gamma: float = 0.0025,
        gradient_clipping: C.str_or_fn = C.use_default,
        update_clipping: C.str_or_fn = C.use_default,
        **kwargs,
    ):
        if len(betas) != 3:
            raise ValueError("AdEMAMix expects betas with three coefficients.")

        defaults = locals()
        defaults.pop("self")
        params = defaults.pop("params")
        defaults.update(defaults.pop("kwargs"))

        if kwargs:
            utils.warn_once(f"Working with uncaptured keyword arguments: {kwargs}")

        super().__init__(params, defaults, foreach, gradient_clipping, update_clipping, fns=(C.update_by_ademamix,))


class UnscaledAdamW(C.BaseOpt):
    def __init__(
        self,
        params,
        lr=0.0025,
        betas=(0.9, 0.99),
        eps=1e-8,
        weight_decay=0,
        warmup_steps=0,
        foreach: bool = True,
        storage_dtype: str = "float32",
        mars: bool = False,
        caution: bool = False,
        mars_gamma: float = 0.0025,
        gradient_clipping: C.str_or_fn = C.use_default,
        update_clipping: C.str_or_fn = C.use_default,
        palm: bool = C.use_default,
        beta2_scale: float = 0.8,
        **kwargs,
    ):
        defaults = locals()
        defaults.pop("self")
        params = defaults.pop("params")
        defaults.update(defaults.pop("kwargs"))

        if kwargs:
            utils.warn_once(f"Working with uncaptured keyword arguments: {kwargs}")

        super().__init__(
            params, defaults, foreach, gradient_clipping, update_clipping, palm, fns=(C.scale_by_unscaled_adam,)
        )


class SUDSAdamW(C.BaseOpt):
    def __init__(
        self,
        params,
        lr=0.0025,
        betas=(0.9, 0.99),
        eps=1e-8,
        weight_decay=0,
        warmup_steps=0,
        foreach: bool = True,
        storage_dtype: str = "float32",
        mars: bool = False,
        caution: bool = False,
        mars_gamma: float = 0.0025,
        gradient_clipping: C.str_or_fn = C.use_default,
        update_clipping: C.str_or_fn = C.use_default,
        palm: bool = C.use_default,
        beta2_scale: float = 0.8,
        precond_lr: float = 1e-2,
        **kwargs,
    ):
        defaults = locals()
        defaults.pop("self")
        params = defaults.pop("params")
        defaults.update(defaults.pop("kwargs"))

        if kwargs:
            utils.warn_once(f"Working with uncaptured keyword arguments: {kwargs}")

        super().__init__(params, defaults, foreach, gradient_clipping, update_clipping, palm, fns=(C.scale_by_suds,))


class Scion(C.BaseOpt):
    def __init__(
        self,
        params,
        lr: float = 0.0025,
        betas: tuple[float, float] = (0.9, 0.99),
        eps: float = 1e-8,
        weight_decay: float = 0,
        warmup_steps: int = 0,
        foreach: bool = True,
        storage_dtype: str = "float32",
        mars: bool = False,
        caution: bool = False,
        mars_gamma: float = 0.0025,
        gradient_clipping: C.str_or_fn = C.use_default,
        update_clipping: C.str_or_fn = C.use_default,
        scale: float = 1.0,
        momentum: Optional[float] = None,
        **kwargs,
    ):
        if lr < 0:
            raise ValueError(f"Invalid learning rate: {lr}")
        if len(betas) == 0 and momentum is None:
            raise ValueError("Scion expects at least one beta or an explicit momentum.")

        beta1 = momentum if momentum is not None else betas[0]
        if not 0 <= beta1 <= 1:
            raise ValueError(f"Invalid momentum value: {beta1}")
        beta2 = betas[1] if len(betas) > 1 else beta1

        defaults = locals()
        defaults.pop("self")
        params = defaults.pop("params")
        defaults.update(defaults.pop("kwargs"))

        if kwargs:
            utils.warn_once(f"Working with uncaptured keyword arguments: {kwargs}")

        defaults["betas"] = (beta1, beta2)
        defaults["scale"] = scale
        defaults.pop("momentum", None)

        super().__init__(
            params, defaults, foreach, gradient_clipping, update_clipping, fns=(C.exp_avg, C.scion_auto_norm)
        )


class ForeachAdamC(C.BaseOpt):
    def __init__(
        self,
        params,
        lr=0.0025,
        betas=(0.9, 0.99),
        eps=1e-8,
        weight_decay=0,
        max_lr: float | None = None,
        warmup_steps=0,
        foreach: bool = True,
        storage_dtype: str = "float32",
        mars: bool = False,
        caution: bool = False,
        mars_gamma: float = 0.0025,
        gradient_clipping: C.str_or_fn = C.use_default,
        update_clipping: C.str_or_fn = C.use_default,
        palm: bool = C.use_default,
        beta2_scale: float = 0.8,
        **kwargs,
    ):
        if max_lr is None:
            utils.warn_once(
                "max_lr was not set. setting it to the current learning rate, under the assumption that it strictly decreases"
            )
            max_lr = lr

        defaults = locals()
        defaults.pop("self")
        params = defaults.pop("params")
        defaults.update(defaults.pop("kwargs"))

        if kwargs:
            utils.warn_once(f"Working with uncaptured keyword arguments: {kwargs}")

        super().__init__(params, defaults, foreach, gradient_clipping, update_clipping, palm, fns=(C.update_by_adamc,))


class ForeachRMSprop(C.BaseOpt):
    """
    Debiased RMSprop (not torch.optim.RMSprop)
    """

    def __init__(
        self,
        params,
        lr=0.0025,
        betas=(0.9, 0.99),
        eps=1e-6,
        weight_decay=0,
        warmup_steps=0,
        r=0.0,
        weight_lr_power=2.0,
        foreach: bool = True,
        storage_dtype: str = "float32",
        mars: bool = False,
        caution: bool = False,
        mars_gamma: float = 0.0025,
        gradient_clipping: C.str_or_fn = C.use_default,
        update_clipping: C.str_or_fn = C.use_default,
        palm: bool = C.use_default,
        beta2_scale: float = 0.8,
        **kwargs,
    ):
        defaults = locals()
        defaults.pop("self")
        params = defaults.pop("params")
        defaults.update(defaults.pop("kwargs"))

        if kwargs:
            utils.warn_once(f"Working with uncaptured keyword arguments: {kwargs}")

        super().__init__(
            params,
            defaults,
            foreach,
            gradient_clipping,
            update_clipping,
            palm,
            fns=(C.scale_by_exp_avg_sq,),
        )


class ForeachSFAdamW(C.ScheduleFree):
    def __init__(
        self,
        params,
        lr=0.0025,
        betas=(0.9, 0.99),
        eps=1e-6,
        weight_decay=0,
        warmup_steps=0,
        r=0.0,
        weight_lr_power=2.0,
        foreach: bool = True,
        storage_dtype: str = "float32",
        mars: bool = False,
        caution: bool = False,
        mars_gamma: float = 0.0025,
        gradient_clipping: C.str_or_fn = C.use_default,
        update_clipping: C.str_or_fn = C.use_default,
        palm: bool = C.use_default,
        beta2_scale: float = 0.8,
        **kwargs,
    ):
        defaults = locals()
        defaults.pop("self")
        params = defaults.pop("params")
        defaults.update(defaults.pop("kwargs"))

        if kwargs:
            utils.warn_once(f"Working with uncaptured keyword arguments: {kwargs}")

        super().__init__(
            params,
            defaults,
            foreach,
            gradient_clipping,
            update_clipping,
            palm,
            fns=(C.scale_by_exp_avg_sq, C.update_by_schedule_free),
        )


class MSAMLaProp(C.MSAM):
    def __init__(
        self,
        params,
        lr=0.0025,
        betas=(0.9, 0.99),
        eps=1e-6,
        weight_decay=0,
        warmup_steps=0,
        r=0.0,
        weight_lr_power=2.0,
        foreach: bool = True,
        storage_dtype: str = "float32",
        mars: bool = False,
        caution: bool = False,
        mars_gamma: float = 0.0025,
        gradient_clipping: C.str_or_fn = C.use_default,
        update_clipping: C.str_or_fn = C.use_default,
        palm: bool = C.use_default,
        beta2_scale: float = 0.8,
        sam_step_size: float = 0.1,
        **kwargs,
    ):
        defaults = locals()
        defaults.pop("self")
        params = defaults.pop("params")
        defaults.update(defaults.pop("kwargs"))

        if kwargs:
            utils.warn_once(f"Working with uncaptured keyword arguments: {kwargs}")

        super().__init__(
            params,
            defaults,
            foreach,
            gradient_clipping,
            update_clipping,
            palm,
            fns=(C.scale_by_exp_avg_sq, C.update_by_msam),
        )


class PaLMForeachSFAdamW(ForeachSFAdamW):
    palm: bool = True


class ForeachADOPT(C.BaseOpt):
    def __init__(
        self,
        params,
        lr=0.0025,
        betas=(0.9, 0.99),
        eps=1e-8,
        weight_decay=0,
        warmup_steps=0,
        foreach: bool = True,
        storage_dtype: str = "float32",
        mars: bool = False,
        caution: bool = False,
        mars_gamma: float = 0.0025,
        gradient_clipping: C.str_or_fn = C.use_default,
        update_clipping: C.str_or_fn = C.use_default,
        palm: bool = C.use_default,
        beta2_scale: float = 0.8,
        **kwargs,
    ):
        defaults = locals()
        defaults.pop("self")
        params = defaults.pop("params")
        defaults.update(defaults.pop("kwargs"))

        if kwargs:
            utils.warn_once(f"Working with uncaptured keyword arguments: {kwargs}")

        super().__init__(params, defaults, foreach, gradient_clipping, update_clipping, palm, fns=(C.update_by_adopt,))


class ForeachMuon(C.BaseOpt):
    def __init__(
        self,
        params,
        lr=0.0025,
        betas=(0.9, 0.99),
        eps=1e-8,
        weight_decay=0,
        warmup_steps=0,
        foreach: bool = True,
        storage_dtype: str = "float32",
        mars: bool = False,
        caution: bool = False,
        mars_gamma: float = 0.0025,
        gradient_clipping: C.str_or_fn = C.use_default,
        update_clipping: C.str_or_fn = C.use_default,
        palm: bool = C.use_default,
        beta2_scale: float = 0.8,
        nesterov: bool = True,
        heavyball_momentum: bool = False,
        **kwargs,
    ):
        defaults = locals()
        defaults.pop("self")
        params = defaults.pop("params")
        defaults.update(defaults.pop("kwargs"))

        if kwargs:
            utils.warn_once(f"Working with uncaptured keyword arguments: {kwargs}")

        if heavyball_momentum:
            if nesterov:
                ema = C.nesterov_momentum
            else:
                ema = C.heavyball_momentum
        elif nesterov:
            ema = C.nesterov_ema
        else:
            ema = C.exp_avg

        super().__init__(
            params,
            defaults,
            foreach,
            gradient_clipping,
            update_clipping,
            palm,
            fns=(ema, C.orthogonalize_update),
        )


class ForeachLaProp(C.BaseOpt):
    def __init__(
        self,
        params,
        lr=0.0025,
        betas=(0.9, 0.99),
        eps=1e-8,
        weight_decay=0,
        warmup_steps=0,
        foreach: bool = True,
        storage_dtype: str = "float32",
        mars: bool = False,
        caution: bool = False,
        mars_gamma: float = 0.0025,
        gradient_clipping: C.str_or_fn = C.use_default,
        update_clipping: C.str_or_fn = C.use_default,
        palm: bool = C.use_default,
        beta2_scale: float = 0.8,
        **kwargs,
    ):
        defaults = locals()
        defaults.pop("self")
        params = defaults.pop("params")
        defaults.update(defaults.pop("kwargs"))

        if kwargs:
            utils.warn_once(f"Working with uncaptured keyword arguments: {kwargs}")

        super().__init__(params, defaults, foreach, gradient_clipping, update_clipping, palm, fns=(C.update_by_laprop,))


class MuonLaProp(C.BaseOpt):
    def __init__(
        self,
        params,
        lr=0.0025,
        betas=(0.9, 0.99),
        eps=1e-8,
        weight_decay=0,
        warmup_steps=0,
        foreach: bool = True,
        storage_dtype: str = "float32",
        mars: bool = False,
        caution: bool = False,
        mars_gamma: float = 0.0025,
        gradient_clipping: C.str_or_fn = C.use_default,
        update_clipping: C.str_or_fn = C.use_default,
        palm: bool = C.use_default,
        beta2_scale: float = 0.8,
        **kwargs,
    ):
        defaults = locals()
        defaults.pop("self")
        params = defaults.pop("params")
        defaults.update(defaults.pop("kwargs"))

        if kwargs:
            utils.warn_once(f"Working with uncaptured keyword arguments: {kwargs}")

        super().__init__(
            params,
            defaults,
            foreach,
            gradient_clipping,
            update_clipping,
            palm,
            fns=(C.scale_by_laprop, C.orthogonalize_update),
        )


class ForeachSOAP(C.BaseOpt):
    """
    ForeachSOAP

    Sources:
        Baseline SOAP:
            SOAP: Improving and Stabilizing Shampoo using Adam
            Nikhil Vyas, Depen Morwani, Rosie Zhao, Itai Shapira, David Brandfonbrener, Lucas Janson, Sham Kakade
            https://arxiv.org/abs/2409.11321
            https://github.com/nikhilvyas/SOAP
    """

    use_precond_schedule: bool = False

    def __init__(
        self,
        params,
        lr: float = 3e-3,
        betas=(0.9, 0.95),
        shampoo_beta: float = 0.95,
        eps: float = 1e-8,
        weight_decay: float = 0.01,
        precondition_frequency: int = 2,
        max_precond_dim: int = 2048,  #
        merge_dims: bool = True,
        precondition_1d: bool = False,
        normalize_grads: bool = False,
        correct_bias: bool = True,
        warmup_steps: int = 0,
        split: bool = False,
        foreach: bool = True,
        mars: bool = False,
        caution: bool = False,
        mars_gamma: float = 0.0025,
        palm: bool = C.use_default,
        precond_scheduler=(1 / 3, 9),
        beta2_scale: float = 0.8,
        use_precond_schedule: bool = C.use_default,
        gradient_clipping: C.str_or_fn = C.use_default,
        update_clipping: C.str_or_fn = C.use_default,
        storage_dtype: str = "float32",
        stochastic_schedule: bool = False,
        precond_grad_accum: bool = False,
        **kwargs,
    ):
        use_precond_schedule = C.default(use_precond_schedule, self.use_precond_schedule)

        defaults = locals()
        defaults.pop("self")
        params = defaults.pop("params")
        defaults.update(defaults.pop("kwargs"))

        if kwargs:
            utils.warn_once(f"Working with uncaptured keyword arguments: {kwargs}")

        if use_precond_schedule:
            del defaults["precondition_frequency"]
            self.precond_schedule = utils.get_soap_precond_schedule(defaults.pop("precond_scheduler"))
        else:
            del defaults["precond_scheduler"]
            self.precond_schedule = 1 / defaults.pop("precondition_frequency")
        super().__init__(
            params,
            defaults,
            foreach,
            gradient_clipping,
            update_clipping,
            palm,  #
            fns=(C.scale_by_soap,),
        )


class ForeachSOAPNAdam(C.BaseOpt):
    use_precond_schedule: bool = False

    def __init__(
        self,
        params,
        lr: float = 3e-3,
        betas=(0.9, 0.999),
        shampoo_beta: float = 0.95,
        eps: float = 1e-8,
        weight_decay: float = 0.01,
        precondition_frequency: int = 2,
        max_precond_dim: int = 2048,
        merge_dims: bool = True,
        precondition_1d: bool = False,
        normalize_grads: bool = False,
        correct_bias: bool = True,
        warmup_steps: int = 0,
        split: bool = False,
        foreach: bool = True,
        mars: bool = False,
        caution: bool = False,
        mars_gamma: float = 0.0025,
        palm: bool = C.use_default,
        precond_scheduler=(1 / 3, 9),
        beta2_scale: float = 0.8,
        use_precond_schedule: bool = C.use_default,
        gradient_clipping: C.str_or_fn = C.use_default,
        update_clipping: C.str_or_fn = C.use_default,
        storage_dtype: str = "float32",
        stochastic_schedule: bool = False,
        precond_grad_accum: bool = False,
        momentum_decay: float = 4e-3,
        decoupled_weight_decay: bool = False,
        **kwargs,
    ):
        use_precond_schedule = C.default(use_precond_schedule, self.use_precond_schedule)

        defaults = locals()
        defaults.pop("self")
        params = defaults.pop("params")
        defaults.update(defaults.pop("kwargs"))

        if kwargs:
            utils.warn_once(f"Working with uncaptured keyword arguments: {kwargs}")

        if use_precond_schedule:
            del defaults["precondition_frequency"]
            self.precond_schedule = utils.get_soap_precond_schedule(defaults.pop("precond_scheduler"))
        else:
            del defaults["precond_scheduler"]
            self.precond_schedule = 1 / defaults.pop("precondition_frequency")
        super().__init__(
            params,
            defaults,
            foreach,
            gradient_clipping,
            update_clipping,
            palm,
            fns=(C.scale_by_soap_nadam,),
        )


class ForeachSOAPAdEMAMix(C.BaseOpt):
    use_precond_schedule: bool = False

    def __init__(
        self,
        params,
        lr: float = 3e-3,
        betas=(0.9, 0.95, 0.999),
        shampoo_beta: float = 0.95,
        eps: float = 1e-8,
        weight_decay: float = 0.01,
        precondition_frequency: int = 2,
        max_precond_dim: int = 2048,
        merge_dims: bool = True,
        precondition_1d: bool = False,
        normalize_grads: bool = False,
        correct_bias: bool = True,
        warmup_steps: int = 0,
        split: bool = False,
        foreach: bool = True,
        mars: bool = False,
        caution: bool = False,
        mars_gamma: float = 0.0025,
        palm: bool = C.use_default,
        precond_scheduler=(1 / 3, 9),
        beta2_scale: float = 0.8,
        use_precond_schedule: bool = C.use_default,
        gradient_clipping: C.str_or_fn = C.use_default,
        update_clipping: C.str_or_fn = C.use_default,
        storage_dtype: str = "float32",
        stochastic_schedule: bool = False,
        precond_grad_accum: bool = False,
        alpha: float = 2.0,
        beta3_warmup: int | None = None,
        alpha_warmup: int | None = None,
        **kwargs,
    ):
        use_precond_schedule = C.default(use_precond_schedule, self.use_precond_schedule)

        defaults = locals()
        defaults.pop("self")
        params = defaults.pop("params")
        defaults.update(defaults.pop("kwargs"))

        if kwargs:
            utils.warn_once(f"Working with uncaptured keyword arguments: {kwargs}")

        if use_precond_schedule:
            del defaults["precondition_frequency"]
            self.precond_schedule = utils.get_soap_precond_schedule(defaults.pop("precond_scheduler"))
        else:
            del defaults["precond_scheduler"]
            self.precond_schedule = 1 / defaults.pop("precondition_frequency")
        super().__init__(
            params,
            defaults,
            foreach,
            gradient_clipping,
            update_clipping,
            palm,
            fns=(C.scale_by_soap_ademamix,),
        )


class ForeachSignLaProp(C.BaseOpt):
    def __init__(
        self,
        params,
        lr=0.0025,
        betas=(0.9, 0.99),
        eps=1e-8,
        weight_decay=0,
        warmup_steps=0,
        foreach: bool = True,
        storage_dtype: str = "float32",
        mars: bool = False,
        caution: bool = False,
        mars_gamma: float = 0.0025,
        gradient_clipping: C.str_or_fn = C.use_default,
        update_clipping: C.str_or_fn = C.use_default,
        palm: bool = C.use_default,
        beta2_scale: float = 0.8,
        **kwargs,
    ):
        defaults = locals()
        defaults.pop("self")
        params = defaults.pop("params")
        defaults.update(defaults.pop("kwargs"))

        if kwargs:
            utils.warn_once(f"Working with uncaptured keyword arguments: {kwargs}")

        super().__init__(
            params,
            defaults,
            foreach,
            gradient_clipping,
            update_clipping,
            palm,
            fns=(C.scale_by_laprop, C.sign),
        )


class ForeachSOLP(C.BaseOpt):
    """
    ForeachSOLP

    Sources:
        Baseline SOAP:
            SOAP: Improving and Stabilizing Shampoo using Adam
            Nikhil Vyas, Depen Morwani, Rosie Zhao, Itai Shapira, David Brandfonbrener, Lucas Janson, Sham Kakade
            https://arxiv.org/abs/2409.11321
            https://github.com/nikhilvyas/SOAP
    """

    use_precond_schedule: bool = False

    def __init__(
        self,
        params,
        lr: float = 3e-3,
        betas=(0.9, 0.95),
        shampoo_beta: float = 0.95,
        eps: float = 1e-8,
        weight_decay: float = 0.01,
        precondition_frequency: int = 2,
        max_precond_dim: int = 2048,  #
        merge_dims: bool = True,
        precondition_1d: bool = False,
        normalize_grads: bool = False,
        correct_bias: bool = True,
        warmup_steps: int = 0,
        split: bool = False,
        foreach: bool = True,
        mars: bool = False,
        caution: bool = False,
        mars_gamma: float = 0.0025,
        palm: bool = C.use_default,
        precond_scheduler=(1 / 3, 9),
        beta2_scale: float = 0.8,
        use_precond_schedule: bool = C.use_default,
        gradient_clipping: C.str_or_fn = C.use_default,
        update_clipping: C.str_or_fn = C.use_default,
        storage_dtype: str = "float32",
        stochastic_schedule: bool = False,
        **kwargs,
    ):
        use_precond_schedule = C.default(use_precond_schedule, self.use_precond_schedule)

        defaults = locals()
        defaults.pop("self")
        params = defaults.pop("params")
        defaults.update(defaults.pop("kwargs"))

        if kwargs:
            utils.warn_once(f"Working with uncaptured keyword arguments: {kwargs}")

        if use_precond_schedule:
            del defaults["precondition_frequency"]
            self.precond_schedule = utils.get_soap_precond_schedule(defaults.pop("precond_scheduler"))
        else:
            del defaults["precond_scheduler"]
            self.precond_schedule = 1 / defaults.pop("precondition_frequency")
        super().__init__(
            params,
            defaults,
            foreach,
            gradient_clipping,
            update_clipping,
            palm,  #
            fns=(C.scale_by_soap_laprop,),
        )


class PaLMForeachSOAP(ForeachSOAP):
    use_precond_schedule: bool = False
    palm: bool = True


class PrecondScheduleForeachSOAP(ForeachSOAP):
    use_precond_schedule: bool = True


class PrecondSchedulePaLMForeachSOAP(ForeachSOAP):
    use_precond_schedule: bool = True
    palm: bool = True


class OrthoLaProp(C.BaseOpt):
    def __init__(
        self,
        params,
        lr=0.0025,
        betas=(0.9, 0.99),
        eps=1e-8,
        weight_decay=0,
        warmup_steps=0,
        foreach: bool = True,
        storage_dtype: str = "float32",
        mars: bool = False,
        caution: bool = False,
        mars_gamma: float = 0.0025,
        gradient_clipping: C.str_or_fn = C.use_default,
        update_clipping: C.str_or_fn = C.use_default,
        palm: bool = C.use_default,
        beta2_scale: float = 0.8,
        **kwargs,
    ):
        defaults = locals()
        defaults.pop("self")
        params = defaults.pop("params")
        defaults.update(defaults.pop("kwargs"))

        if kwargs:
            utils.warn_once(f"Working with uncaptured keyword arguments: {kwargs}")
        super().__init__(
            params,
            defaults,
            foreach,
            gradient_clipping,
            update_clipping,
            palm,
            fns=(C.orthogonalize_grad_to_param, C.scale_by_laprop),
        )


class LaPropOrtho(C.BaseOpt):
    def __init__(
        self,
        params,
        lr=0.0025,
        betas=(0.9, 0.99),
        eps=1e-8,
        weight_decay=0,
        warmup_steps=0,
        foreach: bool = True,
        storage_dtype: str = "float32",
        mars: bool = False,
        caution: bool = False,
        mars_gamma: float = 0.0025,
        gradient_clipping: C.str_or_fn = C.use_default,
        update_clipping: C.str_or_fn = C.use_default,
        palm: bool = C.use_default,
        beta2_scale: float = 0.8,
        **kwargs,
    ):
        defaults = locals()
        defaults.pop("self")
        params = defaults.pop("params")
        defaults.update(defaults.pop("kwargs"))

        if kwargs:
            utils.warn_once(f"Working with uncaptured keyword arguments: {kwargs}")
        super().__init__(
            params,
            defaults,
            foreach,
            gradient_clipping,
            update_clipping,
            palm,
            fns=(C.scale_by_laprop, C.orthogonalize_grad_to_param),
        )


class ForeachPSGDKron(C.BaseOpt):
    """
    Originally from Evan Walters and Omead Pooladzandi, 2024
    Modified under Creative Commons Attribution 4.0 International
    Source available at https://github.com/evanatyourservice/kron_torch/blob/97a2b5ee8a1a4c29e4780bbf6c521e545189eff9/kron_torch/kron.py
    """

    delayed: bool = False
    cached: bool = False
    exp_avg_input: bool = True
    quad: bool = False

    def __init__(
        self,
        params,
        lr=0.001,
        beta=None,
        betas=(0.9, 0.999),
        weight_decay=0.0,
        preconditioner_update_probability=C.use_default,
        max_size_triangular=2048,
        min_ndim_triangular=2,
        memory_save_mode=None,
        momentum_into_precond_update=True,
        warmup_steps: int = 0,
        merge_dims: bool = False,
        split: bool = False,
        store_triu_as_line: bool = True,
        foreach: bool = True,
        q_dtype="float32",
        stochastic_schedule: bool = False,
        storage_dtype: str = "float32",
        mars: bool = False,
        caution: bool = False,
        mars_gamma: float = 0.0025,
        delayed: Optional[bool] = C.use_default,
        cached: Optional[bool] = C.use_default,
        exp_avg_input: Optional[bool] = C.use_default,
        gradient_clipping: C.str_or_fn = C.use_default,
        update_clipping: C.str_or_fn = C.use_default,  #
        adaptive: bool = False,
        ortho_method: Optional[str] = None,  # If None, no orthogonalization
        precond_grad_accum: bool = False,
        lower_bound_beta: float = 0.9,  # 0.0 recovers pre-2.0.0 PSGD
        inverse_free: bool = C.use_default,
        dampening: float = 2**-13,
        precond_update_power_iterations: int = 2,
        # expert parameters
        precond_init_scale=None,
        precond_init_scale_scale: float = 1,
        precond_init_scale_power: Optional[float] = None,
        precond_lr: float = 0.1,
        **kwargs,
    ):
        delayed = C.default(delayed, self.delayed)
        cached = C.default(cached, self.cached)
        exp_avg_input = C.default(exp_avg_input, self.exp_avg_input)
        update_clipping = C.default(update_clipping, utils.trust_region_clip_)
        inverse_free = C.default(inverse_free, self.quad)
        if inverse_free:
            raise ValueError(
                "inverse_free (i.e., PSGD-QUAD) is not supported at the moment. Consider using https://github.com/evanatyourservice/quad_torch"
            )

        defaults = locals()
        defaults.pop("self")
        defaults.update(defaults.pop("kwargs"))

        if kwargs:
            utils.warn_once(f"Working with uncaptured keyword arguments: {kwargs}")

        self.precond_schedule = C.default(
            defaults.pop("preconditioner_update_probability"), utils.precond_update_prob_schedule()
        )
        params = defaults.pop("params")

        super().__init__(
            params,
            defaults,
            foreach,
            gradient_clipping,
            update_clipping,
            False,  #
            fns=(
                *(C.exp_avg,) * exp_avg_input,
                functools.partial(C.scale_by_delayed_psgd if delayed else C.scale_by_psgd, cached=cached),
            ),
        )


class ForeachPurePSGD(ForeachPSGDKron):
    exp_avg_input: bool = False


class ForeachCachedDelayedPSGDKron(ForeachPSGDKron):
    delayed: bool = True
    cached: bool = True


class ForeachCachedPSGDKron(ForeachPSGDKron):
    cached: bool = True


class ForeachDelayedPSGD(ForeachPSGDKron):
    delayed: bool = True


class ForeachCachedNewtonPSGD(ForeachCachedPSGDKron):
    hessian_approx = True


class NewtonHybrid2PSGDKron(ForeachCachedNewtonPSGD):
    hvp_interval = 2


class ForeachPSGDLRA(C.BaseOpt):
    """
    Originally from Evan Walters and Omead Pooladzandi, 2024
    Modified under Creative Commons Attribution 4.0 International
    Source available at https://github.com/evanatyourservice/kron_torch/blob/97a2b5ee8a1a4c29e4780bbf6c521e545189eff9/kron_torch/kron.py
    """

    delayed: bool = False
    exp_avg_input: bool = True

    def __init__(
        self,
        params,
        lr=0.001,
        beta=0.9,
        weight_decay=0.0,
        preconditioner_update_probability=C.use_default,
        momentum_into_precond_update=True,
        rank: Optional[int] = None,
        warmup_steps: int = 0,
        foreach: bool = True,
        q_dtype="float32",
        stochastic_schedule: bool = False,
        storage_dtype: str = "float32",
        mars: bool = False,
        caution: bool = False,
        mars_gamma: float = 0.0025,
        delayed: Optional[bool] = C.use_default,
        exp_avg_input: Optional[bool] = C.use_default,
        gradient_clipping: C.str_or_fn = C.use_default,
        update_clipping: C.str_or_fn = C.use_default,
        eps: float = 1e-8,  #
        precond_grad_accum: bool = False,  # expert parameters
        precond_init_scale=None,
        precond_init_scale_scale: float = 1,
        precond_init_scale_power: Optional[float] = None,
        precond_lr: float = 0.1,
        **kwargs,
    ):
        delayed = C.default(delayed, self.delayed)
        exp_avg_input = C.default(exp_avg_input, self.exp_avg_input)
        update_clipping = C.default(update_clipping, utils.trust_region_clip_)

        defaults = locals()
        defaults.pop("self")
        defaults.update(defaults.pop("kwargs"))

        if kwargs:
            utils.warn_once(f"Working with uncaptured keyword arguments: {kwargs}")

        self.precond_schedule = C.default(
            defaults.pop("preconditioner_update_probability"), utils.precond_update_prob_schedule()
        )
        params = defaults.pop("params")

        if rank is None:
            utils.warn_once(
                f"{rank=}. It will be set to log2(param_count). This requires `params` to be of type list. Currently, {type(params)=}"
            )
            params = list(params)
            defaults["rank"] = round(math.log2(sum(p.numel() for p in params)))
            utils.warn_once(f"rank was set to {defaults['rank']}")

        super().__init__(
            params,
            defaults,
            foreach,
            gradient_clipping,
            update_clipping,
            False,  #
            fns=(*(C.exp_avg,) * exp_avg_input, C.scale_by_delayed_psgd_lra if delayed else C.scale_by_psgd_lra),
        )


class ForeachDelayedPSGDLRA(ForeachPSGDLRA):
    delayed: bool = True


class ForeachNewtonPSGDLRA(ForeachPSGDLRA):
    hessian_approx = True


class NewtonHybrid2PSGDLRA(ForeachNewtonPSGDLRA):
    hvp_interval = 2


class SplitOpt(utils.StatefulOptimizer):
    """
    Delegates different parameter groups to different underlying optimizers.

        opt = SplitOpt([
            {'params': matrices, 'optimizer': Muon, 'lr': 0.02},
            {'params': vectors, 'optimizer': AdamW, 'lr': 0.001},
        ])
    """

    def __init__(self, specs):
        self.optimizers, all_params = [], []
        for spec in specs:
            spec = dict(spec)
            params = list(spec.pop("params"))
            if params:
                self.optimizers.append(spec.pop("optimizer")(params, **spec))
                all_params.extend(params)
        if not self.optimizers:
            raise ValueError("No optimizers created")
        super().__init__(all_params, {}, foreach=True)

    def _step(self, group):
        pass

    def _handle_closure(self, closure):
        return self.optimizers[0]._handle_closure(closure)

    @torch.no_grad()
    def step(self, closure=None):
        loss = self._handle_closure(closure) if closure else None
        for opt in self.optimizers:
            opt.step()
        return loss

    def zero_grad(self, set_to_none: bool = True):
        for opt in self.optimizers:
            opt.zero_grad(set_to_none=set_to_none)

    def state_dict(self):
        return {"optimizers": [opt.state_dict() for opt in self.optimizers]}

    def load_state_dict(self, state_dict):
        for opt, s in zip(self.optimizers, state_dict["optimizers"]):
            opt.load_state_dict(s)


class SAMWrapper(torch.optim.Optimizer):
    def __init__(
        self,
        params,
        wrapped_optimizer: Union[utils.StatefulOptimizer, Type[utils.StatefulOptimizer]] = ForeachAdamW,
        ball: float = 0.1,
    ):
        params = list(params)
        super().__init__(params, {"ball": ball})

        if isinstance(wrapped_optimizer, type):
            if not issubclass(wrapped_optimizer, utils.StatefulOptimizer):
                raise ValueError(f"{wrapped_optimizer.__name__} is not a HeavyBall optimizer")
            wrapped_optimizer = wrapped_optimizer(params)
        elif not isinstance(wrapped_optimizer, utils.StatefulOptimizer):
            raise ValueError(f"{wrapped_optimizer.__class__.__name__} is not a HeavyBall optimizer")

        self.wrapped_optimizer = wrapped_optimizer

    @torch.no_grad()
    def step(self, closure=None):
        if closure is None:
            raise ValueError("SAM requires closure")
        with torch.enable_grad():
            closure()
        old_params = [utils.sam_step(group["params"], group["ball"]) for group in self.param_groups]

        original_handle_closure = self.wrapped_optimizer._handle_closure

        def _handle_closure(closure):
            try:
                _loss = original_handle_closure(closure)
            finally:
                for group, old in zip(self.param_groups, old_params):
                    utils.copy_stochastic_list_(group["params"], old)
            return _loss

        try:
            self.wrapped_optimizer._handle_closure = _handle_closure
            loss = self.wrapped_optimizer.step(closure)
        finally:
            self.wrapped_optimizer._handle_closure = original_handle_closure
        return loss

    def zero_grad(self, set_to_none: bool = True):
        self.wrapped_optimizer.zero_grad(set_to_none=set_to_none)


PalmForEachSoap = PaLMForeachSOAP
PaLMSOAP = PaLMForeachSOAP
PaLMSFAdamW = PaLMForeachSFAdamW
SOAP = ForeachSOAP
SOAPAdEMAMix = ForeachSOAPAdEMAMix
SOAPNAdam = ForeachSOAPNAdam
SFAdamW = ForeachSFAdamW
LaProp = ForeachLaProp
ADOPT = ForeachADOPT
RMSprop = ForeachRMSprop
PrecondScheduleSOAP = PrecondScheduleForeachSOAP
PrecondSchedulePaLMSOAP = PrecondSchedulePaLMForeachSOAP
PSGDKron = ForeachPSGDKron
AdamW = ForeachAdamW
NAdam = ForeachNAdam
PurePSGD = ForeachPurePSGD
DelayedPSGD = ForeachDelayedPSGD
CachedPSGDKron = ForeachCachedPSGDKron
CachedDelayedPSGDKron = ForeachCachedDelayedPSGDKron
Muon = ForeachMuon
SignLaProp = ForeachSignLaProp
DelayedPSGDLRA = ForeachDelayedPSGDLRA
PSGDLRA = ForeachPSGDLRA
NewtonPSGDLRA = ForeachNewtonPSGDLRA
NewtonPSGDKron = ForeachCachedNewtonPSGD

__all__ = [k for k, v in globals().items() if isinstance(v, type) and issubclass(v, torch.optim.Optimizer)]



import pytest
import torch
from lightbench.utils import get_optim
from torch import nn

import heavyball
from heavyball.utils import clean, set_torch


def get_memory():
    clean()
    torch.cuda.synchronize()
    clean()
    torch.cuda.synchronize()
    return torch.cuda.memory_allocated()


expected_memory = {
    "adamw": {"after": 4, "peak": 5.1},
    "soap": {"after": 7, "peak": 14},
    "psgd": {"after": 4, "peak": 11.5},
}


@pytest.mark.parametrize("opt", ["NewtonHybrid2PSGDKron"])
@pytest.mark.parametrize("method", ["qr", "newtonschulz2", "svd", "eigh"])
@pytest.mark.parametrize("size,depth", [(8192, 2), (2048, 16)])
def test_memory(opt, method, size, depth: int, iterations: int = 5, outer_iterations: int = 3):
    if "soap" not in opt.lower() and method != "qr":
        raise pytest.skip("Only SOAP supports `method` argument")
    set_torch()

    for k, v in expected_memory.items():
        if k in opt.lower():
            break
    else:
        raise pytest.skip(f"Opt {opt} not supported")

    opt = getattr(heavyball, opt)
    heavyball.utils.zeroth_power_mode = method

    for i in range(outer_iterations):
        model = nn.Sequential(*[nn.Linear(size, size) for _ in range(depth)]).cuda()
        print(model)
        torch.cuda.reset_peak_memory_stats()
        torch.cuda.reset_max_memory_allocated()
        torch.cuda.reset_max_memory_cached()
        torch.cuda.reset_accumulated_memory_stats()

        model_allocated = get_memory()
        o = get_optim(opt, model.parameters(), lr=1e-3)
        for _ in range(iterations):
            data = torch.randn((1, size), device="cuda").requires_grad_(True)

            def _closure():
                nonlocal model
                loss = (model(data) - data).square().mean()
                loss.backward()
                return loss

            o.step(_closure)

            opt_allocated = get_memory()

        del model, o
        peak = torch.cuda.memory_stats()["allocated_bytes.all.peak"]

        print(f"Peak: {peak / model_allocated:.2f}x | Opt: {opt_allocated / model_allocated:.2f}x")
        if i > 0:
            assert peak / model_allocated < v["peak"]
            assert opt_allocated / model_allocated < v["after"]



import os
from typing import Iterable, List, Sequence

import torch
from hypothesis import given, settings
from hypothesis import strategies as st
from utils import _global_l2_norm, _global_rms_norm, _local_l2_norm, _local_rms_norm

import heavyball
from heavyball.utils import (
    caution,
    global_l2norm_clip,
    global_rmsnorm_clip,
    l2_clip_,
    merge_group,
    rmsnorm_clip_,
    stochastic_add_,
    stochastic_add_divide_,
    stochastic_divide_with_eps_,
    stochastic_multiply_,
)

# Ensure torch.compile stays disabled on CPU-only CI runners.
os.environ.setdefault("TORCH_COMPILE_DISABLE", "1")


heavyball.utils.compile_mode = None

FLOAT_DTYPES: Sequence[torch.dtype] = (torch.float32, torch.bfloat16)
DTYPE_TOLERANCE = {torch.float32: 5e-5, torch.bfloat16: 2e-2}


def _to_float(value: float | torch.Tensor) -> float:
    return float(value.item()) if isinstance(value, torch.Tensor) else float(value)


@st.composite
def _tensor_shape(draw, min_rank: int = 1, max_rank: int = 4, max_side: int = 6) -> Sequence[int]:
    rank = draw(st.integers(min_value=min_rank, max_value=max_rank))
    return tuple(draw(st.integers(min_value=1, max_value=max_side)) for _ in range(rank))


@st.composite
def _tensor_list(
    draw,
    min_len: int = 1,
    max_len: int = 3,
    min_rank: int = 1,
    max_rank: int = 4,
    max_side: int = 6,
) -> List[torch.Tensor]:
    dtype = draw(st.sampled_from(FLOAT_DTYPES))
    length = draw(st.integers(min_value=min_len, max_value=max_len))
    tensors = []
    for _ in range(length):
        shape = draw(_tensor_shape(min_rank=min_rank, max_rank=max_rank, max_side=max_side))
        tensor = torch.randn(shape, dtype=torch.float32)
        tensors.append(tensor.to(dtype))
    return tensors


@st.composite
def _stochastic_inputs(draw) -> tuple[torch.dtype, List[torch.Tensor], List[torch.Tensor]]:
    dtype = draw(st.sampled_from(FLOAT_DTYPES))
    length = draw(st.integers(min_value=1, max_value=3))
    shared_shape = draw(st.booleans())
    tensors: List[torch.Tensor] = []
    if shared_shape:
        shape = draw(_tensor_shape(max_side=5))
        tensors = [torch.randn(shape, dtype=torch.float32).to(dtype) for _ in range(length)]
    else:
        for _ in range(length):
            shape = draw(_tensor_shape(max_side=5))
            tensors.append(torch.randn(shape, dtype=torch.float32).to(dtype))
    partner = [torch.randn_like(tensor, dtype=torch.float32).to(dtype) for tensor in tensors]
    if shared_shape and len(partner) > 1 and draw(st.booleans()):
        partner = [partner[0]]
    return dtype, [tensor.clone() for tensor in tensors], partner


@st.composite
def _caution_inputs(draw) -> tuple[torch.Tensor, torch.Tensor]:
    tensor = draw(_tensor_list(min_len=1, max_len=1, min_rank=1, max_rank=3, max_side=6))[0]
    update = torch.randn_like(tensor, dtype=torch.float32).to(tensor.dtype)
    return tensor, update


def _flatten_tensors(items) -> List[torch.Tensor]:
    if isinstance(items, torch.Tensor):
        return [items]
    flat: List[torch.Tensor] = []
    for item in items:
        flat.extend(_flatten_tensors(item))
    return flat


def _expand_like(reference: Iterable[torch.Tensor], candidate: List[torch.Tensor]) -> List[torch.Tensor]:
    reference = list(reference)
    if len(candidate) == len(reference):
        return [tensor.clone().float() for tensor in candidate]
    assert len(candidate) == 1
    source = candidate[0]
    return [source.clone().float() for _ in reference]


@settings(deadline=None, max_examples=75)
@given(
    tensors=_tensor_list(min_len=1, max_len=3),
    clip_at=st.floats(min_value=1e-3, max_value=10.0, allow_nan=False, allow_infinity=False),
)
def test_local_clipping_respects_thresholds(tensors: List[torch.Tensor], clip_at: float):
    for clip_fn, metric in (
        (l2_clip_, _local_l2_norm),
        (rmsnorm_clip_, _local_rms_norm),
    ):
        clipped = [tensor.clone() for tensor in tensors]
        clip_fn(clipped, clip_at)
        for tensor in clipped:
            tolerance = clip_at * 1e-3 + DTYPE_TOLERANCE.get(tensor.dtype, 1e-5)
            assert metric(tensor) <= clip_at + tolerance


@settings(deadline=None, max_examples=75)
@given(
    tensors=_tensor_list(min_len=1, max_len=3),
    clip_at=st.floats(min_value=1e-3, max_value=10.0, allow_nan=False, allow_infinity=False),
)
def test_global_clipping_bounds_group_norms(tensors: List[torch.Tensor], clip_at: float):
    for clip_fn, metric in (
        (
            global_l2norm_clip,
            _global_l2_norm,
        ),
        (
            global_rmsnorm_clip,
            _global_rms_norm,
        ),
    ):
        clipped = [tensor.clone() for tensor in tensors]
        clip_fn(clipped, clip_at)
        tolerance = clip_at * 1e-3 + DTYPE_TOLERANCE.get(clipped[0].dtype, 1e-5)
        assert metric(clipped) <= clip_at + tolerance


@settings(deadline=None, max_examples=75)
@given(
    data=_stochastic_inputs(),
    alpha=st.one_of(
        st.floats(min_value=-2.0, max_value=2.0, allow_nan=False, allow_infinity=False),
        st.builds(torch.tensor, st.floats(min_value=-2.0, max_value=2.0, allow_nan=False, allow_infinity=False)),
    ),
)
def test_stochastic_add_matches_expected(data, alpha):
    torch.manual_seed(0x172893)
    dtype, tensors, partner = data
    x = [tensor.clone() for tensor in tensors]
    y = [tensor.clone() for tensor in partner]
    stochastic_add_(x, y, alpha)
    alpha_scalar = _to_float(alpha)
    expected_inputs = [tensor.clone().float() for tensor in tensors]
    expected_partner = _expand_like(expected_inputs, partner)
    expected = [xi + yi * alpha_scalar for xi, yi in zip(expected_inputs, expected_partner, strict=True)]
    max_error = max((result.float() - exp).abs().max().item() for result, exp in zip(x, expected, strict=True))
    assert max_error <= DTYPE_TOLERANCE[dtype] + 1e-6


@settings(deadline=None, max_examples=75)
@given(
    data=_stochastic_inputs(),
    alpha=st.floats(min_value=-2.0, max_value=2.0, allow_nan=False, allow_infinity=False),
    divisor=st.one_of(
        st.floats(min_value=0.5, max_value=5.0, allow_nan=False, allow_infinity=False),
        st.builds(torch.tensor, st.floats(min_value=0.5, max_value=5.0, allow_nan=False, allow_infinity=False)),
    ),
)
def test_stochastic_add_divide_matches_expected(data, alpha, divisor):
    torch.manual_seed(0x172893)
    dtype, tensors, partner = data
    x = [tensor.clone() for tensor in tensors]
    y = [tensor.clone() for tensor in partner]
    stochastic_add_divide_(x, y, alpha=alpha, divisor=divisor)
    alpha_scalar = _to_float(alpha)
    divisor_scalar = _to_float(divisor)
    expected_inputs = [tensor.clone().float() for tensor in tensors]
    expected_partner = _expand_like(expected_inputs, partner)
    expected = [
        (xi + yi * alpha_scalar) / divisor_scalar for xi, yi in zip(expected_inputs, expected_partner, strict=True)
    ]
    max_error = max((result.float() - exp).abs().max().item() for result, exp in zip(x, expected, strict=True))
    assert max_error <= DTYPE_TOLERANCE[dtype] + 1e-6


@settings(deadline=None, max_examples=75)
@given(data=_stochastic_inputs())
def test_stochastic_multiply_matches_expected(data):
    torch.manual_seed(0x172893)
    dtype, tensors, partner = data
    x = [tensor.clone() for tensor in tensors]
    y = [tensor.clone() for tensor in partner]
    stochastic_multiply_(x, y)
    expected_inputs = [tensor.clone().float() for tensor in tensors]
    expected_partner = _expand_like(expected_inputs, partner)
    expected = [xi * yi for xi, yi in zip(expected_inputs, expected_partner, strict=True)]
    max_error = max((result.float() - exp).abs().max().item() for result, exp in zip(x, expected, strict=True))
    assert max_error <= DTYPE_TOLERANCE[dtype] + 1e-6


@settings(deadline=None, max_examples=75)
@given(
    data=_stochastic_inputs(),
    eps=st.one_of(
        st.floats(min_value=1e-5, max_value=1.0, allow_nan=False, allow_infinity=False),
        st.builds(torch.tensor, st.floats(min_value=1e-5, max_value=1.0, allow_nan=False, allow_infinity=False)),
    ),
)
def test_stochastic_divide_with_eps_matches_expected(data, eps):
    torch.manual_seed(0x172893)
    dtype, tensors, partner = data
    # Avoid very small denominators to keep the numeric range reasonable.
    partner = [tensor.clamp(min=0.5) for tensor in partner]
    x = [tensor.clone() for tensor in tensors]
    y = [tensor.clone() for tensor in partner]
    stochastic_divide_with_eps_(x, y, eps)
    eps_scalar = _to_float(eps)
    expected_inputs = [tensor.clone().float() for tensor in tensors]
    expected_partner = _expand_like(expected_inputs, partner)
    expected = [xi / (yi + eps_scalar) for xi, yi in zip(expected_inputs, expected_partner, strict=True)]
    max_error = max((result.float() - exp).abs().max().item() for result, exp in zip(x, expected, strict=True))
    assert max_error <= DTYPE_TOLERANCE[dtype] + 1e-6


@settings(deadline=None, max_examples=75)
@given(tensor=_tensor_list(min_len=1, max_len=1, min_rank=2, max_rank=4, max_side=8))
def test_merge_group_passthrough_when_disabled(tensor: List[torch.Tensor]):
    original = tensor[0]
    group = {"merge_dims": False, "max_precond_dim": 8}
    result = merge_group(group, original)
    assert isinstance(result, tuple)
    assert result[0] is original


@settings(deadline=None, max_examples=75)
@given(
    tensor=st.builds(
        lambda tensor: tensor,
        _tensor_list(min_len=1, max_len=1, min_rank=2, max_rank=4, max_side=8),
    ),
    max_dim=st.integers(min_value=2, max_value=12),
    split=st.booleans(),
    use_triangular=st.booleans(),
)
def test_merge_group_preserves_structure(tensor: List[torch.Tensor], max_dim: int, split: bool, use_triangular: bool):
    base = tensor[0]
    group = {"merge_dims": True, "split": split}
    if use_triangular:
        group["max_size_triangular"] = max_dim
    else:
        group["max_precond_dim"] = max_dim
    merged = merge_group(group, base)
    flat = _flatten_tensors(merged)
    assert sum(chunk.numel() for chunk in flat) == base.numel()
    for chunk in flat:
        assert chunk.dtype == base.dtype
        assert chunk.is_contiguous()


@settings(deadline=None, max_examples=75)
@given(_caution_inputs())
def test_caution_masks_disagreeing_directions(data):
    grad_tensor, update_tensor = data
    result = caution(grad_tensor, update_tensor)
    assert result.shape == grad_tensor.shape
    mask = grad_tensor.signbit() ^ update_tensor.signbit()
    scale = mask.numel() / max(mask.numel() - mask.sum().item(), 1)
    expected = update_tensor.clone()
    expected[mask] = 0
    expected = expected * scale
    diff = (result.float() - expected.float()).abs().max().item()
    assert diff <= DTYPE_TOLERANCE[result.dtype] + 1e-6
    positive_alignment = (result * grad_tensor).float().sum().item()
    assert positive_alignment >= -DTYPE_TOLERANCE[result.dtype]



import os

import pytest
import torch
from lightbench.utils import get_optim
from torch import nn

import heavyball
from heavyball.utils import clean, set_torch


def get_memory():
    clean()
    torch.cuda.synchronize()
    clean()
    torch.cuda.synchronize()
    return torch.cuda.memory_allocated()


def _read_int(name: str, default: int, *, minimum: int) -> int:
    raw = os.environ.get(name)
    if raw is None:
        return default
    try:
        return max(minimum, int(raw))
    except ValueError:
        return default


DEFAULT_SIZE = _read_int("HB_FOREACH_TEST_SIZE", 128, minimum=1)
DEFAULT_DEPTH = _read_int("HB_FOREACH_TEST_DEPTH", 16, minimum=1)
DEFAULT_ITERATIONS = _read_int("HB_FOREACH_TEST_ITERATIONS", 64, minimum=1)
DEFAULT_OUTER = _read_int("HB_FOREACH_TEST_OUTER", 1, minimum=1)
DEFAULT_WARMUP = _read_int("HB_FOREACH_TEST_WARMUP", 1, minimum=0)


@pytest.mark.parametrize("opt", heavyball.__all__)
@pytest.mark.parametrize("size,depth", [(DEFAULT_SIZE, DEFAULT_DEPTH)])
def test_foreach(
    opt,
    size,
    depth: int,
    iterations: int = DEFAULT_ITERATIONS,
    outer_iterations: int = DEFAULT_OUTER,
    warmup_runs: int = DEFAULT_WARMUP,
):
    set_torch()

    opt = getattr(heavyball, opt)

    total_runs = warmup_runs + outer_iterations
    assert total_runs >= 1

    peaks = []
    losses = []

    for foreach in [True, False]:
        torch.manual_seed(0x2131290)
        peaks.append([])
        losses.append([])

        for i in range(total_runs):
            clean()
            model = nn.Sequential(*[nn.Linear(size, size) for _ in range(depth)]).cuda()
            clean()

            torch.cuda.reset_peak_memory_stats()
            torch.cuda.reset_max_memory_allocated()
            torch.cuda.reset_max_memory_cached()
            torch.cuda.reset_accumulated_memory_stats()

            clean()
            o = get_optim(opt, model.parameters(), lr=1e-3, foreach=foreach)
            clean()

            for _ in range(iterations):
                loss = model(torch.randn((1, size), device="cuda")).sum()
                loss.backward()
                o.step()
                o.zero_grad()
                losses[-1].append(loss.detach())

            del model, o
            clean()

            peak = torch.cuda.memory_stats()["allocated_bytes.all.peak"]

            if i < warmup_runs:
                continue

            peaks[-1].append(peak)

    if warmup_runs:
        cutoff = warmup_runs * iterations
        losses = [loss_list[cutoff:] for loss_list in losses]

    for p0, p1 in zip(*peaks):
        assert p0 > p1
    for l0, l1 in zip(*losses):  # increase error tolerance for PSGD, as we have different RNGs -> expected differences
        assert torch.allclose(l0, l1, rtol=0.01 if "PSGD" in opt.__class__.__name__ else 1e-5)



import os

os.environ["TORCH_LOGS"] = "+recompiles"

import pytest
import torch
from lightbench.utils import get_optim
from torch import nn
from torch._dynamo import config

import heavyball
from heavyball.utils import clean, set_torch

config.cache_size_limit = 128


@pytest.mark.parametrize("opt", heavyball.__all__)
@pytest.mark.parametrize("size,depth", [(128, 2)])
def test_caution(opt, size, depth: int, iterations: int = 16, outer_iterations: int = 1):
    set_torch()
    opt = getattr(heavyball, opt)
    peaks = []
    losses = []

    for caution in [True, False]:
        torch.manual_seed(0x2131290)
        peaks.append([])
        losses.append([])

        for i in range(outer_iterations):
            model = nn.Sequential(*[nn.Linear(size, size) for _ in range(depth)]).cuda()
            o = get_optim(opt, model.parameters(), lr=1e-5, caution=caution)

            for _ in range(iterations):
                loss = model(torch.randn((1024, size), device="cuda")).square().mean()
                loss.backward()
                o.step()
                o.zero_grad()
                losses[-1].append(loss.detach())

            del model, o
            clean()

    for i, (l0, l1) in enumerate(zip(*losses)):
        print(i, l0.item(), l1.item())
        assert l0.item() <= l1.item() * 1.1



import inspect

import pytest
import torch

import heavyball
from heavyball.utils import StatefulOptimizer

DEVICE = torch.device("cuda" if torch.cuda.is_available() else "cpu")


def _optimizer_params():
    seen = set()
    params = []
    for name in heavyball.__all__:
        if not hasattr(heavyball, name):
            continue
        obj = getattr(heavyball, name)
        if not inspect.isclass(obj):
            continue
        if not issubclass(obj, torch.optim.Optimizer):
            continue
        ident = id(obj)
        if ident in seen:
            continue
        seen.add(ident)
        params.append(pytest.param(name, obj, id=name))
    return params


@pytest.mark.parametrize("opt_name,opt_cls", _optimizer_params())
def test_optimizer_runs_on_cpu(opt_name, opt_cls):
    torch.manual_seed(0xDEADBEEF)

    model = torch.nn.Sequential(
        torch.nn.Linear(8, 16),
        torch.nn.Tanh(),
        torch.nn.Linear(16, 4),
    ).to(DEVICE)

    optimizer = opt_cls(model.parameters())

    initial = [param.detach().clone() for param in model.parameters()]

    def closure():
        optimizer.zero_grad(set_to_none=True)
        data = torch.randn(4, 8, device=DEVICE)
        target = torch.randn(4, 4, device=DEVICE)
        loss = torch.nn.functional.mse_loss(model(data), target)
        loss.backward()
        return loss

    for _ in range(5):
        optimizer.step(closure)

    updated = list(model.parameters())
    deltas = [torch.max(torch.abs(after - before)) for before, after in zip(initial, updated)]
    assert any(delta > 0 for delta in deltas)

    if isinstance(optimizer, StatefulOptimizer):
        # Ensure state dict round-trips without touching CUDA APIs.
        state_dict = optimizer.state_dict()
        clone = opt_cls(model.parameters())
        clone.load_state_dict(state_dict)
        assert clone.state_dict()["state"].keys() == state_dict["state"].keys()



import math
import os

import torch
import torch.testing

import heavyball

# Ensure Torch dynamo stays disabled on CI runners without GPU support.
os.environ.setdefault("TORCH_COMPILE_DISABLE", "1")
heavyball.utils.compile_mode = None


def _alpha_schedule(step: int, alpha: float, warmup: int | None) -> float:
    if not warmup or warmup <= 0:
        return alpha
    if step < warmup:
        return alpha * (step / float(warmup))
    return alpha


def _beta3_schedule(step: int, beta1: float, beta3: float, warmup: int | None, eps: float = 1e-8) -> float:
    if not warmup or warmup <= 0:
        return beta3
    if step >= warmup:
        return beta3

    half_life_start = math.log(0.5) / math.log(beta1 + eps) - 1.0
    half_life_end = math.log(0.5) / math.log(beta3 + eps) - 1.0
    a = step / float(warmup)
    target = (1.0 - a) * half_life_start + a * half_life_end
    beta = math.pow(0.5, 1.0 / (target + 1.0))
    return min(max(beta, 0.0), 1.0 - eps)


def _run_reference_ademamix(
    initial: torch.Tensor,
    grads: list[torch.Tensor],
    betas: tuple[float, float, float],
    lr: float,
    eps: float,
    alpha: float,
    weight_decay: float,
    beta3_warmup: int | None,
    alpha_warmup: int | None,
):
    def _beta_debias(beta: float, step: int) -> float:
        return 1.0 - (1.0 - beta) / (1.0 - beta**step)

    beta1, beta2, beta3 = betas
    param = initial.clone()
    m_fast = torch.zeros_like(initial)
    m_slow = torch.zeros_like(initial)
    v = torch.zeros_like(initial)

    for step, grad in enumerate(grads, start=1):
        grad_tensor = grad.to(param)
        alpha_eff = _alpha_schedule(step, alpha, alpha_warmup)
        beta3_eff = _beta3_schedule(step, beta1, beta3, beta3_warmup)

        beta1_eff = _beta_debias(beta1, step)
        beta2_eff = _beta_debias(beta2, step)

        m_fast = beta1_eff * m_fast + (1.0 - beta1_eff) * grad_tensor
        m_slow = beta3_eff * m_slow + (1.0 - beta3_eff) * grad_tensor
        v = beta2_eff * v + (1.0 - beta2_eff) * grad_tensor.square()

        denom = v.sqrt() + eps
        update = (m_fast + alpha_eff * m_slow) / denom

        param = param * (1.0 - weight_decay * lr) - lr * update

    return param, m_fast, m_slow, v


def test_ademamix_matches_reference_math():
    torch.manual_seed(2024)

    initial = torch.tensor([1.25, -0.75], dtype=torch.float32)
    grads = [
        torch.tensor([0.30, -0.10], dtype=torch.float32),
        torch.tensor([-0.50, 0.20], dtype=torch.float32),
        torch.tensor([0.15, 0.05], dtype=torch.float32),
    ]

    betas = (0.9, 0.999, 0.9999)
    lr = 0.0125
    eps = 1e-8
    alpha = 3.5
    weight_decay = 0.004
    beta3_warmup = 6
    alpha_warmup = 4

    param = torch.nn.Parameter(initial.clone())
    optimizer = heavyball.ForeachAdEMAMix(
        [param],
        lr=lr,
        betas=betas,
        eps=eps,
        weight_decay=weight_decay,
        alpha=alpha,
        beta3_warmup=beta3_warmup,
        alpha_warmup=alpha_warmup,
        foreach=False,
    )

    for grad in grads:
        param.grad = grad.clone()
        optimizer.step()

    expected_param, expected_fast, expected_slow, expected_sq = _run_reference_ademamix(
        initial,
        grads,
        betas,
        lr,
        eps,
        alpha,
        weight_decay,
        beta3_warmup,
        alpha_warmup,
    )

    torch.testing.assert_close(param.detach(), expected_param, atol=1e-6, rtol=1e-5)

    state = optimizer.state[param][0]
    torch.testing.assert_close(state["update_by_ademamix_exp_avg_fast_0"], expected_fast, atol=1e-6, rtol=1e-5)
    torch.testing.assert_close(state["update_by_ademamix_exp_avg_slow_0"], expected_slow, atol=1e-6, rtol=1e-5)
    torch.testing.assert_close(state["update_by_ademamix_exp_avg_sq_0"], expected_sq, atol=1e-6, rtol=1e-5)


def test_soap_ademamix_projects_gradients_into_eigenbasis():
    torch.manual_seed(7)

    param = torch.nn.Parameter(torch.randn(2, 2))
    optimizer = heavyball.ForeachSOAPAdEMAMix([param], lr=0.01, foreach=False)

    # First call initializes the SOAP preconditioner state without applying an update.
    param.grad = torch.randn_like(param)
    optimizer.step()

    state = optimizer.state[param][0]
    Q = [q.clone() if q is not None else None for q in state["scale_by_soap_ademamix_Q_1"]]

    captured: dict[str, list[torch.Tensor] | tuple] = {}
    original = heavyball.utils.ademamix_

    def capture_ademamix(exp_avg_fast, exp_avg_slow, exp_avg_sq, grad, *rest):
        captured["grad"] = [g.clone() for g in grad]
        captured["betas"] = rest[0]
        captured["alpha"] = rest[3]
        captured["beta3_warmup"] = rest[4]
        captured["alpha_warmup"] = rest[5]
        return original(exp_avg_fast, exp_avg_slow, exp_avg_sq, grad, *rest)

    grad_step = torch.randn_like(param)

    heavyball.utils.ademamix_ = capture_ademamix
    try:
        param.grad = grad_step.clone()
        optimizer.step()
    finally:
        heavyball.utils.ademamix_ = original

    assert "grad" in captured, "AdEMAMix inner update was not invoked."

    expected_projected = heavyball.utils.project(grad_step.clone(), Q, False)
    torch.testing.assert_close(captured["grad"][0], expected_projected, atol=1e-6, rtol=1e-5)
    assert captured["betas"] == optimizer.param_groups[0]["betas"]
    assert captured["alpha"] == optimizer.param_groups[0]["alpha"]
    assert captured["beta3_warmup"] == optimizer.param_groups[0].get("beta3_warmup")
    assert captured["alpha_warmup"] == optimizer.param_groups[0].get("alpha_warmup")



import os

import torch

import heavyball.chainable as C
import heavyball.utils

os.environ.setdefault("TORCH_COMPILE_DISABLE", "1")
heavyball.utils.compile_mode = None


def _identity_update(state, group, update, grad, param):
    return update


def test_chain_applies_update_on_cpu():
    param = [torch.nn.Parameter(torch.zeros(2))]
    grad = [torch.ones(2)]
    group = {"lr": 0.1, "caution": False, "weight_decay": 0.0}

    with torch.no_grad():
        C.chain(lambda _: {}, group, grad, param, _identity_update)

    assert torch.allclose(param[0].detach(), torch.full((2,), -0.1))


def test_branch_merges_multiple_paths():
    def double(_, __, update, ___, ____):
        return [u * 2 for u in update]

    def negate(_, __, update, ___, ____):
        return [u * -1 for u in update]

    def merge_fn(outputs):
        return [sum(vals) / len(vals) for vals in zip(*outputs)]

    branch = C.Branch([[double], [negate]], merge_fn)

    update = [torch.ones(2)]
    grad = [torch.ones(2)]
    param = [torch.nn.Parameter(torch.ones(2))]

    result = branch(lambda _: {}, {}, update, grad, param)
    expected = torch.full_like(update[0], 0.5)
    assert torch.allclose(result[0], expected)


def test_set_indices_assigns_transform_ids():
    def base(_, __, update, ___, ____, buffer):
        assert buffer is not None
        return update

    zero_guard = C.ZeroGuard(base, ["buffer"])
    assigned = C.set_indices([zero_guard], retain=False)[0]
    assert assigned.transform_idx == 0

    def state_fn(_x):
        return {}

    group = {"storage_dtype": "float32"}
    update = [torch.ones(1)]
    grad = [torch.ones(1)]
    param = [torch.nn.Parameter(torch.ones(1))]

    assigned(state_fn, group, update, grad, param)



import sys

import numpy as np
import pytest
import torch
from torch._dynamo import config

from heavyball.utils import _lse_mean, divided_root, mean_root, stable_exp

config.cache_size_limit = sys.maxsize
config.accumulated_cache_size_limit = sys.maxsize


def np_mean_pow_root(z_np: np.ndarray, pow_val: float, eps_val: float):
    z_np = np.abs(z_np).clip(min=eps_val) ** pow_val
    mean_pow = np.mean(z_np) ** (1 / pow_val / 2)
    return mean_pow


def _get_tensor(numel, dtype, maxval: float = 10):
    x = torch.randn(numel, dtype=dtype)  # * torch.arange(numel, dtype=dtype)
    x /= x.abs().max()
    x *= maxval
    x = x.clone()
    return x, x.numpy().astype(np.float128)


tolerance = {
    np.float64: {"rtol": 1e-10, "atol": 1e-12},
    np.float32: {"rtol": 1e-5, "atol": 1e-6},
    np.float16: {"rtol": 1e-2, "atol": 1e-3},
}


def _isclose(x, y):
    if isinstance(x, torch.Tensor):
        x = x.cpu().numpy()
    if isinstance(y, torch.Tensor):
        y = y.cpu().numpy()
    if not np.isfinite(x).all():
        assert ~(np.isfinite(x) ^ np.isfinite(y)).any()  # all are nan together
    if np.isfinite(x).any():
        for k, v in tolerance.items():  # numpy doesn't support indexing
            if x.dtype == k:
                tol = v
                break
        else:
            raise ValueError(f"dtype {x.dtype} not supported")
        assert np.allclose(x[np.isfinite(x)], y[np.isfinite(y)], **tol)


@pytest.mark.parametrize("x_val", list(range(-10, 10)))
@pytest.mark.parametrize("dtype", [torch.float16, torch.float32, torch.float64])
def test_stable_exp_scalar(x_val, dtype):
    x = torch.tensor(x_val, dtype=dtype)
    result = stable_exp(x)
    expected = np.exp(x_val) if x_val <= 0 else 1 / np.exp(-x_val)
    _isclose(result.to(x.dtype), expected)


@pytest.mark.parametrize("numel", [2**i for i in range(10)])
@pytest.mark.parametrize("dtype", [torch.float16, torch.float32, torch.float64])
def test_stable_exp_tensor(numel, dtype):
    x, x_np = _get_tensor(numel, dtype)
    result = stable_exp(x)
    expected = np.exp(x_np)
    _isclose(result.to(x.dtype), expected)


@pytest.mark.parametrize("numel", [2**i for i in range(10)])
@pytest.mark.parametrize("pow_val", list(range(1, 16)))
@pytest.mark.parametrize("dtype", [torch.float16, torch.float32, torch.float64])
def test_lse_mean(numel, pow_val, dtype):
    x, x_np = _get_tensor(numel, dtype)
    result = _lse_mean(x, pow_val, 1e-20)
    expected = np.log(np.mean(np.abs(x_np) ** pow_val)) / pow_val / 2
    _isclose(result.to(x.dtype), expected)


@pytest.mark.parametrize("numel", [2**i for i in range(10)])
@pytest.mark.parametrize("pow_val", list(range(1, 16)))
@pytest.mark.parametrize("dtype", [torch.float16, torch.float32, torch.float64])
def test_mean_root(numel, pow_val, dtype):
    x, x_np = _get_tensor(numel, dtype)
    result = mean_root(x, pow_val)
    expected = 1 / (np.mean(np.abs(x_np) ** pow_val) ** (1 / pow_val / 2))
    _isclose(result.to(x.dtype), expected)


@pytest.mark.parametrize("numel", [2**i for i in range(10)])
@pytest.mark.parametrize("pow0_val", list(range(1, 16)))
@pytest.mark.parametrize("pow1_val", list(range(1, 16)))
@pytest.mark.parametrize("dtype", [torch.float16, torch.float32, torch.float64])
def test_divided_root(numel, pow0_val, pow1_val, dtype):
    x, x_np = _get_tensor(numel, dtype)
    y, y_np = _get_tensor(numel, dtype)
    result = divided_root(x, y, pow0_val, pow1_val)
    expected = np_mean_pow_root(x_np, pow0_val, 1e-12) / np_mean_pow_root(y_np, pow1_val, 1e-12)
    _isclose(result.to(x.dtype), expected)



"""Fast CPU-only smoke tests for non-PSGD HeavyBall features."""

from __future__ import annotations

from copy import deepcopy

import pytest
import torch
from torch import nn

import heavyball


def _train_once(optimizer, model: nn.Module, data: torch.Tensor, target: torch.Tensor, steps: int = 3) -> float:
    """Run a few optimization steps and return the final loss."""

    loss = torch.tensor(float("nan"))
    for _ in range(steps):
        optimizer.zero_grad(set_to_none=True)
        prediction = model(data)
        loss = torch.nn.functional.mse_loss(prediction, target)
        loss.backward()
        optimizer.step()
    return loss.item()


def _parameter_drift(model: nn.Module, original: list[torch.Tensor]) -> float:
    current = [param.detach() for param in model.parameters()]
    diffs = [curr - init for curr, init in zip(current, original, strict=True)]
    stacked = torch.cat([diff.reshape(-1) for diff in diffs])
    return stacked.norm().item()


def _make_batch(
    in_features: int = 8, out_features: int = 4, batch: int = 16
) -> tuple[nn.Module, torch.Tensor, torch.Tensor]:
    torch.manual_seed(0x172893)
    model = nn.Sequential(nn.Linear(in_features, out_features), nn.ReLU(), nn.Linear(out_features, out_features))
    data = torch.randn(batch, in_features)
    target = torch.randn(batch, out_features)
    return model, data, target


@pytest.mark.parametrize(
    "opt_name",
    [
        "ForeachSOAP",
        "Muon",
        "ForeachAdamW",
    ],
)
def test_selected_optimizers_run_on_cpu(opt_name: str) -> None:
    model, data, target = _make_batch()
    init = [param.detach().clone() for param in model.parameters()]

    opt_cls = getattr(heavyball, opt_name)
    optimizer = opt_cls(model.parameters(), warmup_steps=0)
    final_loss = _train_once(optimizer, model, data, target, steps=3)

    assert torch.isfinite(torch.tensor(final_loss))
    assert _parameter_drift(model, init) > 0.0


def test_caution_reduces_update_magnitude() -> None:
    baseline_model, data, target = _make_batch()
    cautious_model = deepcopy(baseline_model)

    baseline_init = [param.detach().clone() for param in baseline_model.parameters()]
    cautious_init = [param.detach().clone() for param in cautious_model.parameters()]

    baseline_opt = heavyball.SGD(
        baseline_model.parameters(),
        lr=1e-3,
        caution=False,
    )
    cautious_opt = heavyball.SGD(
        cautious_model.parameters(),
        lr=1e-3,
        caution=True,
    )

    _train_once(baseline_opt, baseline_model, data, target)
    _train_once(cautious_opt, cautious_model, data, target)

    baseline_drift = _parameter_drift(baseline_model, baseline_init)
    cautious_drift = _parameter_drift(cautious_model, cautious_init)

    assert cautious_drift <= baseline_drift * 1.05  # caution should not overshoot compared to baseline


def test_mars_flag_changes_behavior() -> None:
    model_a, data, target = _make_batch()
    model_b = deepcopy(model_a)

    opt_a = heavyball.ForeachAdamW(model_a.parameters(), mars=False, warmup_steps=0)
    opt_b = heavyball.ForeachAdamW(model_b.parameters(), mars=True, warmup_steps=0)

    init = [param.detach().clone() for param in model_a.parameters()]

    _train_once(opt_a, model_a, data, target)
    _train_once(opt_b, model_b, data, target)

    baseline_drift = _parameter_drift(model_a, init)
    mars_drift = _parameter_drift(model_b, init)
    assert baseline_drift > 0.0
    assert mars_drift > 0.0

    deltas = [a.detach() - b.detach() for a, b in zip(model_a.parameters(), model_b.parameters(), strict=True)]
    combined = torch.cat([delta.reshape(-1) for delta in deltas])
    assert combined.norm().item() > 1e-6  # mars path should diverge from baseline


def test_sam_wrapper_requires_closure() -> None:
    model = nn.Linear(4, 2)
    base = heavyball.ForeachAdamW(model.parameters())
    wrapper = heavyball.SAMWrapper(model.parameters(), wrapped_optimizer=base)

    with pytest.raises(ValueError):
        wrapper.step()

    data = torch.randn(8, 4)
    target = torch.randn(8, 2)

    def closure():
        wrapper.zero_grad()
        loss = torch.nn.functional.mse_loss(model(data), target)
        loss.backward()
        return loss

    before = [param.detach().clone() for param in model.parameters()]
    wrapper.step(closure)
    after = [param.detach() for param in model.parameters()]
    diff = torch.cat([(a - b).reshape(-1) for a, b in zip(after, before, strict=True)])
    assert diff.norm().item() > 0.0



import numpy as np
import optuna
import pandas as pd
import torch
from optuna.distributions import FloatDistribution, IntDistribution
from optuna.samplers import RandomSampler
from optuna.trial import TrialState

from heavyball import helpers


def test_bound_to_torch_roundtrip_cpu():
    arr = np.arange(4, dtype=np.float64).reshape(2, 2)
    tensor = helpers.bound_to_torch(arr.tobytes(), arr.shape, "cpu")
    assert torch.allclose(tensor, torch.from_numpy(arr.T))


def test_nextafter_matches_numpy():
    forward = helpers.nextafter(0.5, 1.0)
    backward = helpers.nextafter(1, 0)
    assert forward == np.nextafter(0.5, 1.0)
    assert backward == np.nextafter(1, 0)


def test_untransform_numerical_param_torch_handles_steps():
    dist = FloatDistribution(0.0, 1.0, step=0.1)
    value = torch.tensor(0.46)
    untransformed = helpers._untransform_numerical_param_torch(value, dist, transform_log=False)
    assert torch.isclose(untransformed, torch.tensor(0.5))


def test_simple_api_sampler_suggest_all_returns_expected():
    distributions = {"x": FloatDistribution(0.0, 1.0), "y": IntDistribution(0, 3, step=1)}

    class _Sampler(helpers.SimpleAPIBaseSampler):
        def infer_relative_search_space(self, study, trial):
            return self.search_space

        def sample_relative(self, study, trial, search_space):
            return {}

        def sample_independent(self, study, trial, param_name, param_distribution):
            return trial.params[param_name]

    sampler = _Sampler(distributions)

    class DummyTrial:
        def __init__(self, params):
            self.params = params

        def _suggest(self, name, dist):
            return self.params[name]

    trial = DummyTrial({"x": 0.25, "y": 2})
    suggestions = sampler.suggest_all(trial)
    assert suggestions == {"x": 0.25, "y": 2}


def test_botorch_sampler_sample_relative_smoke(monkeypatch):
    search_space = {"width": FloatDistribution(0.0, 1.0)}
    study = optuna.create_study(direction="minimize", sampler=RandomSampler(seed=0))
    for _ in range(3):
        trial = study.ask()
        width = trial.suggest_float("width", 0.0, 1.0)
        study.tell(trial, width)

    sampler = helpers.BoTorchSampler(search_space, n_startup_trials=1, seed=0, device="cpu")

    def _dummy_candidates(params, values, *_args):
        assert params.shape[1] == 1
        return params.mean(dim=0)

    sampler._candidates_func = _dummy_candidates

    pending = study.ask()
    suggestion = sampler.sample_relative(study, pending, search_space)
    assert "width" in suggestion
    assert 0.0 <= suggestion["width"] <= 1.0


def test_hebo_sampler_observe_and_sample(monkeypatch):
    class DummyHEBO:
        def __init__(self, *_args, **_kwargs):
            self.observed = None

        def suggest(self):
            return pd.DataFrame([{"depth": 0.0}])

        def observe(self, params, values):
            self.observed = (params, values)

    monkeypatch.setattr(helpers, "HEBO", DummyHEBO)

    search_space = {"depth": FloatDistribution(0.0, 1.0)}
    sampler = helpers.HEBOSampler(search_space, seed=1)

    study = optuna.create_study(direction="minimize", sampler=RandomSampler(seed=1))
    trial = study.ask()
    trial.suggest_float("depth", 0.0, 1.0)
    study.tell(trial, 0.2)

    suggestion = sampler.sample_relative(study, study.ask(), search_space)
    assert suggestion["depth"] == 0.0

    completed = study.get_trials(deepcopy=False)[0]
    sampler.after_trial(study, completed, TrialState.COMPLETE, [0.2])
    assert sampler._hebo.observed is not None



import copy
import functools
import math
import random
from collections.abc import Iterable as _Iterable
from typing import Iterable, List, Literal, Optional, Union

import torch
from torch import Tensor

from . import utils

use_default = utils.use_default


def _key_in_state(state, key):
    if isinstance(key, str):
        return key in state
    for k in key:
        if isinstance(k, (tuple, list)):
            continue
        if k not in state:
            return False
    return True


def _inplace_guard_(state, key, template_fn):
    key_not_in_state = not _key_in_state(state, key)
    if key_not_in_state:
        template_fn()
    return key_not_in_state


def _guard_in_state(state, key, template_fn):
    if not _key_in_state(state, key):
        state[key] = template_fn()
    return state[key]


class FunctionTransform:
    def __init__(self, fn, names: list[str] | None = None):
        if names is None:
            names = []
        self.fn = fn
        self.fn_name = self.get_fn().__name__
        self.transform_idx = None
        self.is_initialized = False
        self.names = names

    def _init(self, state: dict, group: dict, update: Tensor, grad: Tensor, param: Tensor, *args, **kwargs):
        raise NotImplementedError

    def _call(self, state, group, update, grad, param, vars, *args, **kwargs):
        raise NotImplementedError

    def __call__(self, state, group, update, grad, param, *args, **kwargs):
        states = [state(p) for p in param]
        skip_update = False
        for st, a in zip(states, zip(update, grad, param, *args)):
            if self.transform_idx not in st.get("is_initialized", set()):
                try:
                    self._init(st, group, *a, **kwargs)
                except SkipUpdate:
                    skip_update = True
                finally:
                    if "is_initialized" not in st:
                        st["is_initialized"] = set()
                    st["is_initialized"].add(self.transform_idx)
        if skip_update:
            raise SkipUpdate from None
        vars = [[st.get(self.val_name(name), None) for st in states] for name in self.names]
        return self._call(state, group, update, grad, param, vars, *args, **kwargs)

    def get_fn(self):
        if utils.hasattr_none(self.fn, "get_fn"):
            return self.fn.get_fn()
        return self.fn

    def val_name(self, name):
        assert self.transform_idx is not None
        return f"{self.fn_name}_{name}_{self.transform_idx}"

    def __repr__(self):
        return f"{self.__class__.__name__}({self.fn}, transform_idx={self.transform_idx})"


class Branch:
    def __init__(self, branches: List[List[callable]], merge_fn: callable):
        self.branches = branches
        self.merge_fn = merge_fn

    def __call__(self, state, group, update, grad, param):
        outputs = []
        for branch in self.branches:
            branch_update = [torch.clone(u, memory_format=torch.preserve_format) for u in update]
            branch_update, skip_update = _inner_chain(state, group, branch_update, grad, param, *branch)
            if skip_update:
                raise ValueError("Branches should not skip updates")
            outputs.append(branch_update)
        return self.merge_fn(outputs)


def _zero_guard(state, key, ref, dtype):
    return _guard_in_state(state, key, lambda: torch.zeros_like(ref, dtype=dtype, memory_format=torch.preserve_format))


def _storage_dtype(group):
    dtype = group.get("storage_dtype", "float32")
    return getattr(torch, dtype)


def _init_mu_product(state, group, update, grad, param, **kwargs):
    dtype = _storage_dtype(group)
    state["mu_product"] = torch.ones((), dtype=dtype, device=param.device)


class ZeroGuard(FunctionTransform):
    def __init__(self, fn, names):
        super().__init__(fn, names)

    def _init(self, state: dict, group: dict, update: Tensor, grad: Tensor, param: Tensor, *args, **kwargs):
        for name in self.names:
            _zero_guard(state, self.val_name(name), param, _storage_dtype(group))

    def _call(self, state, group, update, grad, param, vars, *args, **kwargs):
        return self.fn(state, group, update, grad, param, *args, *vars, **kwargs)


class PrecondGradAccumGuard(FunctionTransform):
    def __init__(self, fn):
        super().__init__(fn, ["precond_grad_accum"])
        self.steps_taken = 0
        self.pass_through = None

    def _accum(self, state, new):
        self.steps_taken += 1
        utils.stochastic_add_(state, new)

    def _reset(self, state):
        if self.steps_taken != 0:
            self.steps_taken = 0
            utils.zero_(state)

    def _init(self, state: dict, group: dict, update: Tensor, grad: Tensor, param: Tensor, *args, **kwargs):
        if self.pass_through is None:
            self.pass_through = not group.get("precond_grad_accum", False)
        if self.pass_through is False:
            for name in self.names:
                _zero_guard(state, self.val_name(name), param, _storage_dtype(group))

    def _call(self, state, group, update, grad, param, vars, *args, **kwargs):
        base_grad = update if group.get("momentum_into_precond_update", True) else grad
        if self.pass_through:
            return self.fn(state, group, update, grad, param, *args, base_grad, **kwargs)

        (vars,) = vars
        if group["is_preconditioning"]:
            if self.steps_taken:
                self._accum(vars, base_grad)
                utils.stochastic_multiply_(vars, 1 / self.steps_taken)
            else:
                vars = base_grad
        else:
            self._accum(vars, base_grad)
            vars = base_grad
        try:
            out = self.fn(state, group, update, grad, param, *args, vars, **kwargs)
        finally:
            if group["is_preconditioning"]:
                self._reset(vars)

        return out


class CopyGuard(FunctionTransform):
    def __init__(self, fn, index, names):
        super().__init__(fn, names)
        self.index = index

    def _init(self, state: dict, group: dict, update: Tensor, grad: Tensor, param: Tensor, *args, **kwargs):
        val = [update, grad, param, *args][self.index]
        for name in self.names:
            state[self.val_name(name)] = torch.clone(val)

    def _call(self, state, group, update, grad, param, vars, *args, **kwargs):
        return self.fn(state, group, update, grad, param, *args, *vars, **kwargs)


class GeneralGuard(FunctionTransform):
    def __init__(self, fn, names, init_fn, skip_first: bool = True):
        super().__init__(fn, names)
        self.init_fn = init_fn
        self.skip_first = skip_first
        self.named_to_anonymous = None
        self.anonymous_to_named = None

    def _map(self, state_fn, param, mapping):
        for p in param:
            state = state_fn(p)
            for name, mapped in mapping.items():
                if mapped in state:
                    raise ValueError(f"Name {name} already mapped to {mapped}")
                if name in state:
                    state[mapped] = state.pop(name)

    def _init(self, state: dict, group: dict, update: Tensor, grad: Tensor, param: Tensor, *args, **kwargs):
        self.init_fn(state, group, update, grad, param, **kwargs)
        for name in self.names:
            state[self.val_name(name)] = state.pop(name, None)
        if self.skip_first:
            raise SkipUpdate from None

    def _call(self, state, group, update, grad, param, vars, *args, **kwargs):
        return self.fn(state, group, update, grad, param, *args, *vars, **kwargs)


class NoState(FunctionTransform):
    def __call__(self, state, group, update, grad, param, *args, **kwargs):
        return self.fn(group, update, grad, param, *args, **kwargs)


class NoStateNoForeach(FunctionTransform):
    def __call__(self, state, group, update, grad, param, *args, **kwargs):
        updates = []
        skip_update = False
        for a in zip(update, grad, param, *args):
            try:
                updates.append(self.fn(group, *a, **kwargs))
            except SkipUpdate:
                skip_update = True
                pass
        if skip_update:
            raise SkipUpdate from None
        return updates


class SqueezeGrad(FunctionTransform):
    def __call__(self, state, group, update, grad, param, *args, **kwargs):
        original_shapes = [u.shape for u in update]
        update = [u.squeeze() if u.numel() > 1 else u.view(-1) for u in update]
        grad = [x.view_as(u) for x, u in zip(grad, update)]
        param = [x.view_as(u) for x, u in zip(param, update)]
        args = list(args)
        for i, a in enumerate(args):
            if isinstance(a, (list, tuple)) and isinstance(a[0], Tensor):
                args[i] = [x.view_as(u) for x, u in zip(a, update)]
        for k, a in kwargs.items():
            if isinstance(a, (list, tuple)) and isinstance(a[0], Tensor):
                kwargs[k] = [x.view_as(u) for x, u in zip(a, update)]
        out = self.fn(state, group, update, grad, param, *args, **kwargs)
        return [o.view(s) for o, s in zip(out, original_shapes)]


def zero_guard(*names):
    return functools.partial(ZeroGuard, names=names)


def copy_guard(index, *names):
    return functools.partial(CopyGuard, index=index, names=names)


def general_guard(*names, init_fn, skip_first: bool = True):
    return functools.partial(GeneralGuard, names=names, init_fn=init_fn, skip_first=skip_first)


def no_state(fn):
    return NoState(fn)


def no_state_no_foreach(fn):
    return NoStateNoForeach(fn)


class SkipUpdate(ValueError):
    pass


@zero_guard("exp_avg")
@no_state
def exp_avg(group, update, grad, param, exp_avg):
    return utils.scale_by_exp_avg_(exp_avg, update, utils.beta_debias(utils.get_beta1(group), group["step"]))


@copy_guard(2, "init")
@no_state
def weight_decay_to_init(group, update, grad, param, init):
    utils.stochastic_lerp_(param, init, group["weight_decay_to_ema"] * group["lr"])
    return update


def identity(state, group, update, grad, param):
    return update


@zero_guard("exp_avg")
@no_state
def weight_decay_to_ema(group, update, grad, param, exp_avg):
    utils.weight_decay_to_ema_(
        param,
        exp_avg,
        utils.beta_debias(group["ema_beta"], group["step"]),
        group["weight_decay_to_ema"] * group["lr"],
    )
    return update


@no_state
def cautious_weight_decay(group, update, grad, param):
    utils.cautious_weight_decay_(param, update, group["cautious_weight_decay"] * group["lr"])
    return update


@zero_guard("exp_avg")
@no_state
def l1_weight_decay_to_ema(group, update, grad, param, exp_avg):
    utils.l1_weight_decay_to_ema_(
        param,
        exp_avg,
        utils.beta_debias(group["ema_beta"], group["step"]),
        group["weight_decay_to_ema"] * group["lr"],
    )
    return update


@zero_guard("exp_avg_sq")
@no_state
def scale_by_exp_avg_sq(group, update, grad, param, exp_avg_sq):
    return utils.scale_by_exp_avg_sq_(
        exp_avg_sq, update, utils.beta_debias(utils.get_beta2(group), group["step"]), group["eps"]
    )


@zero_guard("exp_avg", "exp_avg_sq")
@no_state
def scale_by_adam(group, update, grad, param, exp_avg, exp_avg_sq):
    return utils.adam_(
        exp_avg,
        exp_avg_sq,
        update,
        utils.get_beta1(group),
        utils.get_beta2(group),
        group["step"],  #
        group["eps"],
    )


@zero_guard("exp_avg", "exp_avg_sq")
@no_state
def update_by_adam(group, update, grad, param, exp_avg, exp_avg_sq):
    utils.fused_adam_(
        param,
        exp_avg,
        exp_avg_sq,
        update,
        grad,
        utils.get_beta1(group),
        utils.get_beta2(group),
        group["step"],
        group["lr"],
        group["eps"],
        group["weight_decay"],
        group["caution"],
    )
    raise SkipUpdate from None


@zero_guard("exp_avg", "exp_avg_sq")
@general_guard("mu_product", init_fn=_init_mu_product, skip_first=False)
@no_state
def scale_by_nadam(group, update, grad, param, exp_avg, exp_avg_sq, mu_product):
    utils.nadam_(
        param,
        exp_avg,
        exp_avg_sq,
        mu_product,
        update,
        utils.get_beta1(group),
        utils.get_beta2(group),
        group["step"],
        group["momentum_decay"],
        group["eps"],
        group["weight_decay"],
        group.get("decoupled_weight_decay", False),
    )
    return update


@zero_guard("exp_avg", "exp_avg_sq")
@general_guard("mu_product", init_fn=_init_mu_product, skip_first=False)
@no_state
def update_by_nadam(group, update, grad, param, exp_avg, exp_avg_sq, mu_product):
    utils.fused_nadam_(
        param,
        exp_avg,
        exp_avg_sq,
        mu_product,
        update,
        grad,
        utils.get_beta1(group),
        utils.get_beta2(group),
        group["step"],
        group["lr"],
        group["eps"],
        group["momentum_decay"],
        group["weight_decay"],
        group.get("decoupled_weight_decay", False),
        group["caution"],
    )
    raise SkipUpdate from None


@zero_guard("exp_avg", "exp_avg_sq")
@no_state
def update_by_adamc(group, update, grad, param, exp_avg, exp_avg_sq):
    utils.fused_adam_(
        param,
        exp_avg,
        exp_avg_sq,
        update,
        grad,
        utils.get_beta1(group),
        utils.get_beta2(group),
        group["step"],
        group["lr"],
        group["eps"],
        group["lr"] * group["weight_decay"] / group["max_lr"],
        group["caution"],
    )
    raise SkipUpdate from None


@zero_guard("exp_avg_fast", "exp_avg_slow", "exp_avg_sq")
@no_state
def scale_by_ademamix(group, update, grad, param, exp_avg_fast, exp_avg_slow, exp_avg_sq):
    return utils.ademamix_(
        exp_avg_fast,
        exp_avg_slow,
        exp_avg_sq,
        update,
        group["betas"],
        group["step"],
        group["eps"],
        group["alpha"],
        group.get("beta3_warmup"),
        group.get("alpha_warmup"),
    )


@zero_guard("exp_avg_fast", "exp_avg_slow", "exp_avg_sq")
@no_state
def update_by_ademamix(group, update, grad, param, exp_avg_fast, exp_avg_slow, exp_avg_sq):
    utils.fused_ademamix_(
        param,
        exp_avg_fast,
        exp_avg_slow,
        exp_avg_sq,
        update,
        grad,
        group["betas"],
        group["step"],
        group["lr"],
        group["eps"],
        group["weight_decay"],
        group["alpha"],
        group["caution"],
        group.get("beta3_warmup"),
        group.get("alpha_warmup"),
    )
    raise SkipUpdate from None


@zero_guard("exp_avg", "exp_avg_sq")
@no_state
def scale_by_laprop(group, update, grad, param, exp_avg, exp_avg_sq):
    return utils.laprop_(exp_avg, exp_avg_sq, update, utils.get_beta1(group), utils.get_beta2(group), group["step"])


@zero_guard("exp_avg", "exp_avg_sq")
@no_state
def update_by_laprop(group, update, grad, param, exp_avg, exp_avg_sq):
    utils.fused_laprop_(
        param,
        exp_avg,
        exp_avg_sq,
        update,
        grad,
        utils.get_beta1(group),
        utils.get_beta2(group),
        group["step"],
        group["lr"],
        group["weight_decay"],
        group["caution"],
    )
    raise SkipUpdate from None


@no_state
def orthogonalize_grad_to_param(group, update, grad, param):
    return utils.orthogonalize_grad_to_param(param, update, group["eps"])


@copy_guard(2, "z")
@no_state
def update_by_schedule_free(group, update, grad, param, z):
    group["weight_sum"] = utils.schedule_free_(
        group["lr"],
        group["weight_lr_power"],
        group.get("weight_sum", 0),
        utils.get_beta1(group),
        param,
        z,
        update,
        grad,
        group["caution"],
        group["r"],
        group["step"],
        group["weight_decay"],
    )
    raise SkipUpdate from None


@copy_guard(2, "z")
@zero_guard("exp_avg")
@no_state
def update_by_msam(group, update, grad, param, z, exp_avg):
    utils.msam_(
        group["lr"],
        utils.beta_debias(utils.get_beta1(group), group["step"]),
        param,
        z,
        update,
        grad,
        exp_avg,
        group["caution"],
        group["weight_decay"],
        group["sam_step_size"],
    )
    raise SkipUpdate from None


@zero_guard("exp_avg", "exp_avg_sq")
@no_state
def update_by_adopt(group, update, grad, param, exp_avg, exp_avg_sq):
    if group["step"] == 1:
        utils.scale_by_exp_avg_sq_(exp_avg_sq, update, 0, group["eps"])
        raise SkipUpdate from None

    if group["step"] == 2:
        update = utils.promote(update)
        easq = utils.promote(exp_avg_sq)
        [utils.set_(ea, u / easq_.sqrt().clamp_(min=group["eps"])) for ea, u, easq_ in zip(exp_avg, update, easq)]
        utils.scale_by_exp_avg_sq_(
            exp_avg_sq,
            update,
            utils.beta_debias(utils.get_beta2(group), group["step"]),
            group["eps"],
        )
        raise SkipUpdate from None

    utils.fused_adopt_(
        param,
        update,
        grad,
        exp_avg_sq,
        exp_avg,
        utils.get_beta1(group),
        utils.get_beta2(group),
        group["step"] - 2,
        group["lr"],
        group["eps"],
        group["weight_decay"],
        group["caution"],
    )
    raise SkipUpdate from None


@zero_guard("exp_avg", "exp_avg_sq", "fisher_approx")
@no_state_no_foreach
def scale_by_suds(group, update, grad, param, exp_avg, exp_avg_sq, fisher_approx):
    if group["step"] == 1:
        utils.copy_stochastic_(fisher_approx, update / update.norm().clamp(min=1e-8))
        raise SkipUpdate from None

    precond_update, w = utils.eigvecs_product_rank1(update.flatten(), fisher_approx.flatten().to(update.dtype))
    precond_update = utils.adam_(
        exp_avg,
        exp_avg_sq,
        precond_update.view_as(exp_avg),
        utils.get_beta1(group),
        utils.get_beta2(group),
        group["step"] - 1,
    )[0]
    precond_update, _ = utils.eigvecs_product_rank1(precond_update.flatten(), fisher_approx.flatten(), w)

    new_approx = utils.oja_update(fisher_approx.flatten().to(update.dtype), update.flatten(), group["precond_lr"])
    new_approx = new_approx.view_as(fisher_approx)
    utils.copy_stochastic_(fisher_approx, new_approx)
    return precond_update


@zero_guard("exp_avg", "exp_avg_sq")
@no_state
def scale_by_unscaled_adam(group, update, grad, param, exp_avg, exp_avg_sq):
    update = utils.unscaled_adam_(
        exp_avg,
        exp_avg_sq,
        update,
        utils.get_beta1(group),
        utils.get_beta2(group),
        group["step"],
    )
    return update


@zero_guard("exp_avg", "exp_avg_sq")
@no_state
def scale_by_adopt(group, update, grad, param, exp_avg, exp_avg_sq):
    if group["step"] == 1:
        utils.scale_by_exp_avg_sq_(exp_avg_sq, update, 0, group["eps"])
        raise SkipUpdate from None

    if group["step"] == 2:
        update = utils.promote(update)
        easq = utils.promote(exp_avg_sq)
        [utils.set_(ea, u / easq_.sqrt().clamp_(min=group["eps"])) for ea, u, easq_ in zip(exp_avg, update, easq)]
        utils.scale_by_exp_avg_sq_(
            exp_avg_sq,
            update,
            utils.beta_debias(utils.get_beta2(group), group["step"]),
            group["eps"],
        )
        raise SkipUpdate from None

    return utils.adopt(
        update,
        exp_avg_sq,
        exp_avg,
        utils.get_beta1(group),
        utils.get_beta2(group),
        group["step"] - 2,
    )


def _init_psgd_kron(state, group, update, grad, param, cached: bool = False, prob: Optional[callable] = None):
    Q = utils.init_Q_exprs(
        grad,
        group["precond_init_scale"],
        group["precond_init_scale_scale"],
        group["precond_init_scale_power"],
        group["max_size_triangular"],
        group["min_ndim_triangular"],
        group["memory_save_mode"],
        getattr(param, "hessian_vector", None),
        getattr(param, "vector", None),
        dtype=getattr(torch, group["q_dtype"]),
    )
    state["Q"] = utils.triu_to_line(Q) if group["store_triu_as_line"] else Q
    state["running_lower_bound"] = [torch.zeros((1,), device=q.device, dtype=torch.float64) for q in Q]
    state["step"] = torch.zeros((), device=param.device, dtype=torch.float64)  # torch casts int to float in ckpt load
    if group["adaptive"]:
        state["velocity"] = [torch.zeros((), device=q.device, dtype=q.dtype) for q in Q]
    if not cached:
        return

    state["Q_cache"] = [torch.empty_like(q) for q in Q]


def _init_psgd_lra(state, group, update, grad, param, cached: bool = False, prob: Optional[callable] = None):
    state["U"], state["V"], state["d"] = utils.init_lra(
        grad,
        group["param_count"],
        group["precond_init_scale"],
        group["precond_init_scale_scale"],
        group["precond_init_scale_power"],
        group["rank"],
        getattr(param, "hessian_vector", None),
        getattr(param, "vector", None),
        dtype=getattr(torch, group["q_dtype"]),
    )


def precond_schedule(group, prob: Union[callable, float, None] = None, name: str = "cumulative_prob"):
    step = group["step"]
    if "precondition_frequency" in group:
        return step > 0 and step % group["precondition_frequency"] == 0
    if isinstance(step, torch.Tensor):
        utils.warn_once("Preconditioner schedule is not supported with torch.Tensor step.")
        rng = random.Random(0x172381)
    else:
        rng = random.Random(0x172381 ^ step)
    if "precond_scheduler" in group:
        return utils.precond_schedule(step, group["precond_scheduler"], rng)
    if prob is not None:
        return utils.psgd_should_update(group, prob, rng, name=name)
    raise ValueError("No preconditioner update schedule specified.")


@no_state_no_foreach
def orthogonalize_update(group, update, grad, param, scale_mode: str = "scale"):  # explore scale_mode="graft"
    if update.dim() < 2:
        return update
    original_shape = update.shape
    # doing it this way, as tmp and update are not guaranteed to share memory address or layout
    tmp = update.flatten(1, -1)
    utils.inplace_orthogonal_(tmp, out=tmp, scale_mode=scale_mode)
    return tmp.reshape(original_shape)


@zero_guard("momentum")
@no_state
def nesterov_momentum(group, updates, grads, params, momentum):
    return utils.nesterov_momentum(momentum, updates, utils.get_beta1(group))


@zero_guard("momentum")
@no_state
def nesterov_ema(group, updates, grads, params, momentum):  # equivalent to Grokfast
    return utils.nesterov_ema(momentum, updates, utils.get_beta1(group))


def _store_std(state, group, update, grad, param):
    state["init_std"] = torch.std(param)


@general_guard("init_std", init_fn=_store_std, skip_first=False)
@no_state
def mup_approx(group, updates, grads, params, init_std):
    _updates = [(u, i) for u, i in zip(updates, init_std) if u.ndim > 1]
    _updates, _init_std = zip(*_updates)
    utils.stochastic_multiply_(_updates, _init_std)
    return updates


def _init_delta(state, group, update, grad, param, log_space: bool):
    val = group["initial_d"]
    state["delta"] = torch.full((), math.log(val) if log_space else val, dtype=param.dtype, device=param.device)


def _init_full_delta(state, group, update, grad, param, log_space: bool):
    val = group["initial_d"]
    state["delta"] = torch.full_like(param, math.log(val) if log_space else val)


@zero_guard("state")
@general_guard("delta", init_fn=functools.partial(_init_delta, log_space=False), skip_first=False)
@no_state
def scale_by_d_adaptation(group, update, grad, param, state, delta):
    utils.d_adaptation(grad, update, state, delta)
    return update


@zero_guard("state")
@general_guard("delta", init_fn=functools.partial(_init_delta, log_space=True), skip_first=False)
@no_state
def scale_by_lr_adaptation(group, update, grad, param, state, delta):
    utils.lr_adaptation(grad, update, state, delta, group["lr_lr"])
    return update


@zero_guard("state")
@general_guard("delta", init_fn=functools.partial(_init_full_delta, log_space=True), skip_first=False)
@no_state
def scale_by_pointwise_lr_adaptation(group, update, grad, param, state, delta):
    utils.pointwise_lr_adaptation(grad, update, state, delta, group["lr_lr"])
    return update


@zero_guard("momentum")
@no_state
def heavyball_momentum(group, updates, grads, params, momentum):
    return utils.heavyball_momentum(momentum, updates, utils.get_beta1(group))


def _init_scion_state(state, group, update, grad, param):
    state["scion_state"] = {"initialized": False}


@general_guard("scion_state", init_fn=_init_scion_state, skip_first=False)
@no_state
def scion_auto_norm(group, update, grad, param, scion_state):
    scale = group.get("scale", 1.0)
    for ctx, p in zip(scion_state, param):
        if not ctx["initialized"]:
            utils.scion_auto_init_param_(p, scale)
            ctx["initialized"] = True
    return utils.scion_auto_lmo_(update, scale)


def _init_soap(state, group, update, grad, param):
    utils.init_preconditioner(grad, state, group["max_precond_dim"], group["precondition_1d"])


def _apply_soap_preconditioner(group, update, Q, GG, *references):
    for upd, q, gg, *ref in zip(update, Q, GG, *references):
        utils.update_preconditioner(
            utils.promote(upd),
            q,
            gg,
            ref,
            group["max_precond_dim"],
            group["precondition_1d"],
            utils.beta_debias(group["shampoo_beta"], group["step"]),
            group["is_preconditioning"],
        )


@zero_guard("exp_avg", "exp_avg_sq")
@general_guard("Q", "GG", init_fn=_init_soap)
@no_state
def scale_by_soap(group, update, grad, param, exp_avg, exp_avg_sq, Q, GG):
    grad_projected = [utils.project(utils.promote(u), q, False) for u, q in zip(update, Q)]
    precond = utils.adam_(
        exp_avg,
        exp_avg_sq,
        grad_projected,
        utils.get_beta1(group),
        utils.get_beta2(group),
        group["step"] - 1,
        group["eps"],
    )
    precond = [utils.project(p, q, True) for p, q in zip(precond, Q)]
    _apply_soap_preconditioner(group, update, Q, GG, exp_avg)
    return precond


@zero_guard("exp_avg", "exp_avg_sq")
@general_guard("mu_product", init_fn=_init_mu_product, skip_first=False)
@general_guard("Q", "GG", init_fn=_init_soap)
@no_state
def scale_by_soap_nadam(group, update, grad, param, exp_avg, exp_avg_sq, mu_product, Q, GG):
    grad_projected = [utils.project(utils.promote(u), q, False) for u, q in zip(update, Q)]
    precond = utils.nadam_(
        grad_projected,
        exp_avg,
        exp_avg_sq,
        mu_product,
        grad_projected,
        utils.get_beta1(group),
        utils.get_beta2(group),
        group["step"] - 1,
        group["momentum_decay"],
        group["eps"],
        0.0,
        False,
    )
    precond = [utils.project(p, q, True) for p, q in zip(precond, Q)]
    _apply_soap_preconditioner(group, update, Q, GG, exp_avg)
    return precond


@zero_guard("exp_avg", "exp_avg_sq")
@general_guard("Q", "GG", init_fn=_init_soap)
@no_state
def scale_by_soap_laprop(group, update, grad, param, exp_avg, exp_avg_sq, Q, GG):
    grad_projected = [utils.project(utils.promote(u), q, False) for u, q in zip(update, Q)]
    precond = utils.laprop_(
        exp_avg,
        exp_avg_sq,
        grad_projected,
        utils.get_beta1(group),
        utils.get_beta2(group),
        group["step"] - 1,
    )
    precond = [utils.project(p, q, True) for p, q in zip(precond, Q)]
    _apply_soap_preconditioner(group, update, Q, GG, exp_avg)
    return precond


@zero_guard("exp_avg_fast", "exp_avg_slow", "exp_avg_sq")
@general_guard("Q", "GG", init_fn=_init_soap)
@no_state
def scale_by_soap_ademamix(group, update, grad, param, exp_avg_fast, exp_avg_slow, exp_avg_sq, Q, GG):
    grad_projected = [utils.project(utils.promote(u), q, False) for u, q in zip(update, Q)]
    precond = utils.ademamix_(
        exp_avg_fast,
        exp_avg_slow,
        exp_avg_sq,
        grad_projected,
        group["betas"],
        group["step"] - 1,
        group["eps"],
        group["alpha"],
        group.get("beta3_warmup"),
        group.get("alpha_warmup"),
    )
    precond = [utils.project(p, q, True) for p, q in zip(precond, Q)]
    _apply_soap_preconditioner(group, update, Q, GG, exp_avg_slow, exp_avg_fast)
    return precond


def _update_psgd_precond(
    cached, Q_cache, group, param, grad, Q, velocity, running_lower_bound, step, prob: Optional[callable] = None
) -> Optional[Tensor]:
    if prob is None:
        prob = utils.precond_update_prob_schedule()

    if not group["is_preconditioning"]:
        return

    if utils.hasattr_none(param, "vector"):
        vector, hessian_vector = param.vector, param.hessian_vector
        del param.vector
        del param.hessian_vector
    elif group["inverse_free"]:
        vector, hessian_vector = None, grad
    else:
        vector, hessian_vector = utils.dampen_grad(grad, group["dampening"])

    precond = utils.psgd_update_precond(
        hessian_vector,
        group["precond_lr"],
        Q,
        group["store_triu_as_line"],
        velocity,
        utils.get_beta2(group),
        group["ortho_method"],
        vector,
        running_lower_bound,
        group["lower_bound_beta"],
        group["precond_update_power_iterations"],
    )
    del vector, hessian_vector

    if isinstance(prob, float):
        float_prob = prob
    else:
        prob_step = group.get(f"cumulative_prob_{id(Q)}_prob_step", 1)
        float_prob = prob(prob_step)
        group[f"cumulative_prob_{id(Q)}_prob_step"] = prob_step + 1
    group["is_cached"] = should_use_cache = cached and float_prob < 0.5

    if precond is not None:
        return precond
    if not should_use_cache or not cached:
        return None  # caching adds extra ops and is not worth the overhead when we precondition at every step

    for c_, q_ in zip(Q_cache, utils.line_to_triu(Q, group["inverse_free"]) if group["store_triu_as_line"] else Q):
        if q_.ndim == 2:
            torch.matmul(q_.T, q_, out=c_)
        else:
            torch.mul(q_, q_, out=c_)
    return None


def _cached_psgd_precond_grad(group, update, Q, Q_cache, grad):
    kwargs = {"ea": update, "caution": group["caution"], "grad": grad}
    if group.get("is_cached", False):
        out = utils.precond_grad_cached_(cached_q=Q_cache, **kwargs)
    else:
        out = utils.psgd_precond_grad(
            preconds=Q, store_triu_as_line=group["store_triu_as_line"], symmetric_output=group["inverse_free"], **kwargs
        )
    group["caution"] = False  # we already cautioned here - shouldn't do it again
    return out


def _fused_cached_psgd_precond_grad(group, grad, param, update, Q, Q_cache):
    kwargs = {
        "ea": update,
        "caution": group["caution"],
        "grad": grad,
        "param": param,
        "lr": group["lr"],
        "decay": group["weight_decay"],
    }
    if group.get("is_cached", False):
        utils.fused_precond_grad_cached_(cached_q=Q_cache, **kwargs)
    else:
        utils.fused_psgd_precond_grad(
            preconds=Q, store_triu_as_line=group["store_triu_as_line"], symmetric_output=group["inverse_free"], **kwargs
        )


def _update_lra(
    group, U: List[Tensor], V: List[Tensor], d: List[Tensor], params: List[Tensor], grads: List[Tensor], delayed: bool
):
    if not group["is_preconditioning"]:
        return utils.multi_flatten((U, 1), (V, 1), (d, 0))

    if utils.hasattr_none(params[0], "hessian_vector"):
        vector = utils.flatten([p.vector for p in params])
        hessian_vector = utils.flatten([p.hessian_vector for p in params])
        for p in params:
            del p.vector
            del p.hessian_vector
    else:
        vector, hessian_vector = utils.dampen_multiple(grads)
    precond_step = group["precond_step"] = group.get("precond_step", -1) + 1
    return utils.update_lra_precond_(
        U, V, d, vector, hessian_vector, group["eps"], group["precond_lr"], delayed, bool(precond_step % 2)
    )


@SqueezeGrad
@PrecondGradAccumGuard
@general_guard("U", "V", "d", init_fn=_init_psgd_lra, skip_first=False)
@no_state
def scale_by_psgd_lra(group, update, grad, param, update_to_precond, U, V, d):
    u, v, d = _update_lra(group, U, V, d, param, update_to_precond, False)
    return utils.extract_from_flat_update(param, utils.lra_precond(u, v, d, utils.flatten(update)))


@SqueezeGrad
@PrecondGradAccumGuard
@general_guard("U", "V", "d", init_fn=_init_psgd_lra, skip_first=False)
@no_state
def update_by_psgd_lra(group, update, grad, param, update_to_precond, U, V, d):
    u, v, d = _update_lra(group, U, V, d, param, update_to_precond, False)
    utils.apply_lra_update(param, update, u, v, d, group["lr"], group["weight_decay"], group["caution"], grad)
    raise SkipUpdate from None


@SqueezeGrad
@PrecondGradAccumGuard
@general_guard("U", "V", "d", init_fn=_init_psgd_lra, skip_first=False)
@no_state
def scale_by_delayed_psgd_lra(group, update, grad, param, update_to_precond, U, V, d):
    u, v, d = _update_lra(group, U, V, d, param, update_to_precond, True)
    return utils.extract_from_flat_update(param, utils.lra_precond(u, v, d, utils.flatten(update)))


@SqueezeGrad
@PrecondGradAccumGuard
@general_guard("U", "V", "d", init_fn=_init_psgd_lra, skip_first=False)
@no_state
def update_by_delayed_psgd_lra(group, update, grad, param, update_to_precond, U, V, d):
    u, v, d = _update_lra(group, U, V, d, param, update_to_precond, True)
    utils.apply_lra_update(param, update, u, v, d, group["lr"], group["weight_decay"], group["caution"], grad)
    raise SkipUpdate from None


@SqueezeGrad
@PrecondGradAccumGuard
@general_guard("Q", "Q_cache", "velocity", "running_lower_bound", "step", init_fn=_init_psgd_kron, skip_first=False)
@no_state_no_foreach
def scale_by_psgd(
    group,
    update,
    grad,
    param,
    update_to_precond,
    Q,
    Q_cache,
    velocity: Optional[List[Tensor]],
    running_lower_bound: List[Tensor],
    step: Tensor,
    cached: bool = False,
    prob: Optional[callable] = None,
):
    _update_psgd_precond(cached, Q_cache, group, param, update_to_precond, Q, velocity, running_lower_bound, step, prob)
    return _cached_psgd_precond_grad(group, update, Q, Q_cache, grad)


@SqueezeGrad
@PrecondGradAccumGuard
@general_guard("Q", "Q_cache", "velocity", "running_lower_bound", "step", init_fn=_init_psgd_kron, skip_first=False)
@no_state_no_foreach
def scale_by_delayed_psgd(
    group,
    update,
    grad,
    param,
    update_to_precond,
    Q,
    Q_cache,
    velocity: Optional[List[Tensor]],
    running_lower_bound: List[Tensor],
    step: Tensor,
    cached: bool = False,
    prob: Optional[callable] = None,
):
    if group.get("inverse_free", False):
        precond = None
    else:
        precond = _cached_psgd_precond_grad(group, update, Q, Q_cache, grad)
    new = _update_psgd_precond(
        cached, Q_cache, group, param, update_to_precond, Q, velocity, running_lower_bound, step, prob
    )
    return new if precond is None else precond


@SqueezeGrad
@PrecondGradAccumGuard
@general_guard("Q", "Q_cache", "velocity", "running_lower_bound", "step", init_fn=_init_psgd_kron, skip_first=False)
@no_state_no_foreach
def update_by_psgd(
    group,
    update,
    grad,
    param,
    update_to_precond,
    Q,
    Q_cache,
    velocity: Optional[List[Tensor]],
    running_lower_bound: List[Tensor],
    step: Tensor,
    cached: bool = False,
    prob: Optional[callable] = None,
):
    _update_psgd_precond(cached, Q_cache, group, param, update_to_precond, Q, velocity, running_lower_bound, step, prob)
    _fused_cached_psgd_precond_grad(group, update, param, update, Q, Q_cache)
    raise SkipUpdate from None


@no_state
def sign(group, update, grad, param, graft: bool = True):
    return utils.sign_(update, graft)


@no_state
def global_clip(group, update, grad, param, clip_fn: Optional[callable] = None):
    assert clip_fn is not None
    return clip_fn(update)


@SqueezeGrad
@PrecondGradAccumGuard
@general_guard("Q", "Q_cache", "velocity", "running_lower_bound", "step", init_fn=_init_psgd_kron, skip_first=False)
@no_state_no_foreach
def update_by_delayed_psgd(
    group,
    update,
    grad,
    param,
    update_to_precond,
    Q,
    Q_cache,
    velocity: Optional[List[Tensor]],
    running_lower_bound: List[Tensor],
    step: Tensor,
    cached: bool = False,
    prob: Optional[callable] = None,
):
    _fused_cached_psgd_precond_grad(group, update, param, update, Q, Q_cache)
    _update_psgd_precond(cached, Q_cache, group, param, update_to_precond, Q, velocity, running_lower_bound, step, prob)
    raise SkipUpdate from None


def palm_beta2(state, group, update, grad, param):
    beta2 = 1 - group["step"] ** -group["beta2_scale"]
    group["betas"] = (utils.get_beta1(group), beta2)
    return update


def apply_to_idx(fn, idx):
    def _fn(state, group, update, grad, param):
        args = [state, group, update, grad, param]
        return fn(args[idx])

    return _fn


def _inner_chain(state, group, update, grad, param, *fns):
    skip_update = False
    for fn in fns:
        try:
            update = fn(state, group, update, grad, param)
        except SkipUpdate:
            skip_update = True
            continue
        if update is None:
            break
    return update, skip_update


def chain(state: Union[callable, dict], group, grad, param, *fns):
    update = [torch.clone(g, memory_format=torch.preserve_format) for g in grad]
    update, skip_update = _inner_chain(state, group, update, grad, param, *fns)
    if not skip_update and update is not None:
        utils.update_param_(param, update, group["lr"], group["weight_decay"], caution=group["caution"], grad=grad)


def set_indices(fns: Iterable[callable], retain: bool = True, offset: int = 0):
    if retain and offset:
        raise ValueError("offset cannot be retained")

    def _walk(obj):
        stack = [obj]
        while stack:
            cur = stack.pop()
            if isinstance(cur, FunctionTransform):
                yield cur
                stack.append(cur.fn)
            elif isinstance(cur, functools.partial):
                stack.append(cur.func)
            elif isinstance(cur, Branch):
                for branch in cur.branches:
                    stack.extend(branch)
            elif isinstance(cur, _Iterable) and not isinstance(cur, (str, bytes, bytearray)):
                stack.extend(cur)

    if retain:
        offset = max((ft.transform_idx for ft in _walk(fns) if ft.transform_idx is not None), default=-1) + 1

    new_fns = [copy.deepcopy(fn) for fn in fns]
    for ft in _walk(new_fns):
        if not retain or ft.transform_idx is None:
            ft.transform_idx, offset = offset, offset + 1

    return new_fns


class ChainOpt(utils.StatefulOptimizer):
    promote: bool = False
    global_defaults = {
        "caution": False,
        "lr": 1,
        "warmup_steps": 0,
        "weight_decay": 0,
        "eps": 1e-8,
    }

    def __init__(self, params, defaults, foreach: bool, *fns):
        base = self.global_defaults.copy()
        base.update({k: v for k, v in defaults.items() if v is not use_default})
        super().__init__(params, base, foreach)
        self.fns = fns

    @property
    def fns(self):
        return self._fns

    @fns.setter
    def fns(self, value):
        self._fns = value
        self._set_indices(retain=True)

    def _set_indices(self, retain=True):
        self._fns = set_indices(self.fns, retain)

    def _step(self, group):
        if "base_lr" not in group:
            group["base_lr"] = group["lr"]
        if "prev_lr" in group and group["prev_lr"] != group["lr"]:
            utils.warn_once(
                f"Learning rate changed between steps. This is an experimental feature and "
                f"only supported with foreach=True (currently foreach={group['foreach']})."
            )
            group["base_lr"] = group["lr"]

        caution = group["caution"]

        vals = list(self.split_p_and_g_in_group(group, should_promote=self.promote, beta1=utils.get_beta1(group)))

        if not vals:
            return
        p, g = zip(*vals)

        for param in p:
            state = self.state_(param)
            if "step" in state:
                step = state["step"]
            elif self.compile_step:
                step = utils.scalar_guard(0, param)
            else:
                step = 0
            break

        group["step"] = state["step"] = step = step + 1
        group["prev_lr"] = group["lr"] = group["base_lr"] * step / max(step, group["warmup_steps"] + 1)

        if not group["foreach"] or len(p) == 1:
            for param, grad in zip(p, g):
                chain(self.state_, group, [grad], [param], *self.fns)
                group["caution"] = caution
        else:
            chain(self.state_, group, g, p, *self.fns)

        group["caution"] = caution
        group["lr"] = group["prev_lr"]
        group["step"] = None


str_or_fn = Union[str, callable, None, Literal[use_default]]


def _get_clip_fn(name: str_or_fn, default_val: str_or_fn):
    name = default(name, default_val)
    if callable(name):
        return name
    elif name not in (
        "l2_clip_",
        "rmsnorm_clip_",
        "trust_region_clip_",
        "a_law_compress",
        "mu_law_compress",
    ):
        raise ValueError(f"Clipping function {name} not found")
    return getattr(utils, name)


def default(a, b):
    return b if a is use_default else a


# not supported: update_by_schedule_free, scale_by_soap, scale_by_exp_avg_sq
_scale_to_update_map = {
    scale_by_delayed_psgd.get_fn(): update_by_delayed_psgd,  #
    scale_by_psgd.get_fn(): update_by_psgd,  #
    scale_by_psgd_lra.get_fn(): update_by_psgd_lra,  #
    scale_by_delayed_psgd_lra.get_fn(): update_by_delayed_psgd_lra,  #
    scale_by_adam.get_fn(): update_by_adam,  #
    scale_by_nadam.get_fn(): update_by_nadam,  #
    scale_by_laprop.get_fn(): update_by_laprop,  #
    scale_by_adopt.get_fn(): update_by_adopt,  #
    scale_by_ademamix.get_fn(): update_by_ademamix,  #
}
_scale_to_update_map_inv = {
    update_by_delayed_psgd.get_fn(): scale_by_delayed_psgd,  #
    update_by_psgd.get_fn(): scale_by_psgd,  #
    update_by_psgd_lra.get_fn(): scale_by_psgd_lra,  #
    update_by_delayed_psgd_lra.get_fn(): scale_by_delayed_psgd_lra,  #
    update_by_adam.get_fn(): scale_by_adam,  #
    update_by_nadam.get_fn(): scale_by_nadam,  #
    update_by_laprop.get_fn(): scale_by_laprop,  #
    update_by_adopt.get_fn(): scale_by_adopt,  #
    update_by_ademamix.get_fn(): scale_by_ademamix,  #
}


class BaseOpt(ChainOpt):
    """
    Base Optimizer

    compile_step: bool = False
    Whether to change some internals to try to make the optimizer compilable
    This does not compile the step by itself and breaks some optimizers loudly (e.g. SOAP)

    promote: bool = False
    Whether to promote the gradients to fp32 before applying the optimizer
    Improves update quality for low-precision parameters, but increases costs
    Compiling the optimizer step would reduce memory and compute. Alternatively, `foreach=False` decreases memory at the cost of runtime

    gradient_clipping: str_or_fn = None
    The function to use for clipping the incoming gradients, before any other transformations.
    This is syntactic sugar, equivalent to manually passing the function as the first element of the optimizer chain.

    update_clipping: str_or_fn = None
    The function to use for clipping the outgoing updates before applying them, after all other transformations.
    This will turn off fused updates.
    This is syntactic sugar, equivalent to manually passing the function as the last element of the optimizer chain.

    """

    gradient_clipping: str_or_fn = None
    update_clipping: str_or_fn = None
    palm: bool = False
    auto_fuse: bool = True

    def __init__(
        self,
        params,
        defaults,
        foreach: bool = True,
        gradient_clipping: str_or_fn = None,
        update_clipping: str_or_fn = None,
        palm: bool = use_default,
        fns: Iterable[callable] = (),
        compile_step: bool = use_default,
        promote: bool = use_default,
    ):
        if not fns:
            raise ValueError("No functions provided. If that's on purpose (SGD-like), use `identity`")

        args, kwargs = None, None
        fns = tuple(fns)
        fn = fns[-1]
        if isinstance(fn, functools.partial):
            fn, args, kwargs = fn.func, fn.args, fn.keywords
        if isinstance(fn, FunctionTransform):
            fn = fn.get_fn()

        if default(update_clipping, self.update_clipping) is None:
            if self.auto_fuse:
                if fn in _scale_to_update_map:
                    fn = _scale_to_update_map[fn]
                    if args is not None:
                        fn = functools.partial(fn, *args, **kwargs)
                    fns = tuple(fns)[:-1] + (fn,)
        elif fn in _scale_to_update_map_inv:
            if not self.auto_fuse:
                raise ValueError(
                    "update_clipping is currently not compatible with update_by_* functions. "
                    "Manually select scale_by_* functions or set auto_fuse=True."
                )
            fn = _scale_to_update_map_inv[fn]
            if args is not None:
                fn = functools.partial(fn, *args, **kwargs)
            fns = tuple(fns)[:-1] + (fn,)

        self.compile_step = default(compile_step, self.compile_step)
        self.promote = default(promote, self.promote)
        if default(palm, self.palm):
            fns = (palm_beta2,) + fns
        if default(gradient_clipping, self.gradient_clipping) is not None:
            fns = (apply_to_idx(gradient_clipping, 2),) + fns
        if default(update_clipping, self.update_clipping) is not None:
            fns = fns + (apply_to_idx(update_clipping, 2),)

        super().__init__(params, defaults, foreach, *fns)


class ScheduleFree(BaseOpt):
    def eval(self):
        for group in self.param_groups:
            group["train_mode"] = train_mode = not group.get("train_mode")
            beta1 = utils.get_beta1(group)
            if beta1 > 0 and not train_mode:
                for p in group["params"]:
                    state = self.state_(p)
                    if "z" in state:
                        # Set p.data to x
                        z = utils.promote(state["z"])
                        p32 = utils.promote(p.data)
                        p32.lerp_(end=z, weight=1 - 1 / beta1)
                        utils.copy_stochastic_(p.data, p32)

    def train(self):
        for group in self.param_groups:
            group["train_mode"] = train_mode = not group.get("train_mode")
            beta1 = utils.get_beta1(group)
            if beta1 > 0 and train_mode:
                for p in group["params"]:
                    state = self.state_(p)
                    if "z" in state:
                        z = utils.promote(state["z"])
                        p32 = utils.promote(p.data)
                        p32.lerp_(end=z, weight=1 - beta1)
                        utils.copy_stochastic_(p.data, p32)


class MSAM(BaseOpt):
    def eval(self):
        for group in self.param_groups:
            group["train_mode"] = train_mode = not group.get("train_mode")
            if not train_mode:
                for p in group["params"]:
                    state = self.state_(p)
                    if "z" in state:
                        p_copy = p.data.clone()
                        utils.copy_stochastic_(p.data, state["z"])
                        utils.copy_stochastic_(state["z"], p_copy)

    def train(self):
        for group in self.param_groups:
            group["train_mode"] = train_mode = not group.get("train_mode")
            if train_mode:
                for p in group["params"]:
                    state = self.state_(p)
                    if "z" in state:
                        p_copy = p.data.clone()
                        utils.copy_stochastic_(p.data, state["z"])
                        utils.copy_stochastic_(state["z"], p_copy)
